package com.agentfactory.afapl.interpreter;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;
import java.util.Set;

import com.agentfactory.clf.interpreter.AbstractLanguageAgent;
import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.plans.IPlanStep;
import com.agentfactory.clf.reasoner.Bindings;
import com.agentfactory.clf.reasoner.IQueryEngine;
import com.agentfactory.clf.reasoner.IQueryable;
import com.agentfactory.clf.reasoner.QueryResult;
import com.agentfactory.clf.util.LogicException;

public class AbstractAFAPLAgent extends AbstractLanguageAgent implements IQueryable {
    public static final String TYPE = "AFAPL";

    /*********************************************************************
     * Core Interpreter Data Structures
     *********************************************************************/
    protected List<GoalHandler> goals;
    protected List<IMPLIES> inferences;
    protected IQueryEngine queryEngine;
    protected List<Commitment> commitments;
    protected List<Plan> plans;
    protected List<CommitmentRule> rules;
    protected Set<String> eventFilter;
    protected BeliefManager beliefManager;

    public AbstractAFAPLAgent( String name ) {
        super( name );

        goals = new LinkedList<GoalHandler>();
        inferences = new LinkedList<IMPLIES>();
        beliefManager = new BeliefManager();
        queryEngine = new AFAPLResolutionBasedQueryEngine();
        queryEngine.addSource( this );
        queryEngine.addSource( beliefManager );
        eventFilter = new HashSet<String>();
        rules = new LinkedList<CommitmentRule>();
        commitments = new LinkedList<Commitment>();

        installBasicAPI();
    }

    @Override
    public void execute() {
        this.retrieveNewMessages();
        beliefManager.wipeBeliefs( "percept" );
        if ( this.getIterations() == 1 )
            beliefManager.wipeBeliefs( "initial" );
        this.senseEnvironment();

        beliefManager.manage();
        this.getOutbox().clear();

        int index = 0;
        while ( index < goals.size() ) {
            GoalHandler handler = goals.get( index );
            if ( handler.handle( this ) ) {
                index++;
            }
            else {
                goals.remove( index );
            }
        }

        for ( CommitmentRule rule : rules ) {
            QueryResult result = queryEngine.queryAll( rule.condition );
            if ( result.found ) {
                if ( result.bindings.isEmpty() ) {
                    commitments.add( new Commitment( rule ) );
                }
                else {
                    for ( Bindings binding : result.bindings ) {
                        commitments.add( new Commitment( rule.apply( binding ) ) );
                    }
                }
            }
        }

        // Part C: Handle Existing Commitments (if any exist)
        if ( !commitments.isEmpty() ) {
            Iterator<Commitment> it = commitments.iterator();
            while (it.hasNext()) {
                Commitment commitment = it.next();
                if (commitment.isSuspended()) continue;
                
                IPlanStep planStep = commitment.getNextActivity();
                if ( planStep == null ) {
                    // No Action => Commitment is completed
                    it.remove();
                }
                else {
                    planStep.handle( this, commitment );
                    if ( commitment.isFailed() ) {
                        if ( !commitment.rollback() ) {
                            it.remove();
                        }
                    }
                }
            }
        }

        // Signal the end of the iteration...
        this.endOfIteration();
    }

    public QueryResult queryAll( IFormula formula ) {
        return queryEngine.queryAll( formula );
    }

    @Override
    /**
     * Implement response to lack of an action
     */
    public void noSuchAction( Predicate activity ) {
        String a = CoreUtilities.presenter.toString( activity );
        System.out.println( "Unknown Action Call: " + a );
        addBelief( (Predicate) Utilities.factory.createFormula( "unknownActivity(" + a + ")[source(percept)]" ) );
    }

    public List<Predicate> getBeliefs() {
        return beliefManager.beliefs();
    }

    public List<Commitment> getCommitments() {
        return commitments;
    }

    public List<CommitmentRule> getRules() {
        List<CommitmentRule> list = new LinkedList<CommitmentRule>();
        list.addAll( rules );
        return list;
    }

    @Override
    public String getType() {
        return TYPE;
    }

    @Override
    public void initialise( String data ) {
        try {
            addBelief( (Predicate) Utilities.factory.createFormula( data + "[source(initial)]" ) );
        }
        catch ( LogicException le ) {
            System.out.println( "[" + getName() + "] Initialisation Error: " + le.getMessage() );
        }
    }

    @Override
    public void update( Observable arg0, Object arg1 ) {
    // TODO Auto-generated method stub
    }

    protected void addInference( IMPLIES inference ) {
        inferences.add( inference );
    }

    public List<IMPLIES> getInferences() {
        return inferences;
    }

    public void addBelief( Predicate belief ) {
        beliefManager.adoptBelief( belief, null );
    }

    public void addBelief( Predicate belief, Object source ) {
        beliefManager.adoptBelief( belief, source );
    }

    public void removeBelief( Predicate belief ) {
        beliefManager.dropBelief( belief, null );
    }

    public void removeBelief( Predicate belief, Object source ) {
        beliefManager.dropBelief( belief, source );
    }

    protected void addCommitmentRule( CommitmentRule rule ) {
        rules.add( rule );
    }

    @Override
    public boolean hasType( String type ) {
        return true;
    }

    @Override
    public List<IFormula> query( IFormula query ) {
        List<IFormula> list = new LinkedList<IFormula>();
        if ( Predicate.class.isInstance( query ) ) {
            Predicate predicate = (Predicate) query;
            for ( IMPLIES inference : inferences ) {
                if ( inference.result().matches( predicate ) ) {
                    list.add( inference );
                }
            }
        }
        if ( Goal.class.isInstance( query ) ) {
            Goal goal = (Goal) query;
            for ( GoalHandler handler : this.goals ) {
                if ( handler.goal().matches( goal ) ) {
                    list.add( handler.goal() );
                }
            }
        }
        return list;
    }

    public List<Plan> getPlans() {
        return plans;
    }

    public GoalHandler addGoal( Goal goal ) {
        GoalHandler handler = new GoalHandler( goal );
        goals.add( handler );
        return handler;
    }

    public void removeGoal( Goal goal ) {
        goals.remove( goal );
    }

    public List<GoalHandler> getGoals() {
        return goals;
    }
}