package com.agentfactory.afapl.interpreter;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.clf.lang.AND;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.Unifier;
import com.agentfactory.clf.plans.IPlanStep;
import com.agentfactory.clf.reasoner.Bindings;
import com.agentfactory.clf.reasoner.QueryResult;

public class GoalHandler {
	public static final int NEW								= 0;
	public static final int READY							= 1;
	public static final int ACTIVE							= 2;
	public static final int SUSPENDED						= 3;
	public static final int SUCCEEDED						= 4;
	public static final int FAILED   						= 5;
	
	private Goal goal;
	private Commitment source;
	private List<Plan> options;
	private Plan current;
	private Commitment commitment;
	private int state;
	
	public GoalHandler(Goal goal) {
		this.goal = goal;
		state = NEW;
	}
	
	public void setSource(Commitment source) {
		this.source = source;
	}
	
	public int state() {
		return state;
	}
	
	public Goal goal() {
		return goal;
	}
	
	public boolean handle(AbstractAFAPLAgent agent) {
		// Check if the goal is satisfied - if it is, move to SUCCEEDED state
		QueryResult result = agent.queryAll(goal.formula());
		if (result.found) {
			state = SUCCEEDED;
		}
		
		switch (state) {
		case NEW:
			options = new LinkedList<Plan>();
			
			for (Plan plan : agent.getPlans()) {
	            Bindings set = getBindingForPostCondition(plan.post);
	
	            if (set != null) {
	                result = agent.queryAll(plan.pre.apply(set));
	                if (result.found) {
	                    if (result.bindings.isEmpty()) {
	                    	options.add(plan.apply(set));
	                    } else {
	                        options.add(plan.apply(set).apply(result.bindings.get(0)));
	                    }
	                }
	            }
	        }

			state = READY;
		case READY:
			if (options.isEmpty()) {
				state = FAILED;
			} else {
				current = options.remove(0);
				if (source == null) {
					// Create a new commitment
					commitment = new Commitment(current);
					agent.commitments.add(commitment);
				} else {
					// Push the goal steps onto the existing commitment...
					List<IPlanStep> list = new LinkedList<IPlanStep>();
					list.add(current.plan);
					source.addPlan(list);
				}
				state = ACTIVE;
			}
			break;
		case ACTIVE:
			if (commitment != null) {
				if (commitment.isFailed()) {
					state = FAILED;
				} else if (commitment.isSucceeded()) {
	                state = READY;
				}
			}
			break;
		case FAILED:
            agent.addBelief((Predicate) Utilities.factory.createFormula("failedGoal(" + Utilities.presenter.toString(goal.formula()) + ")"));
            commitment = null;
            break;
		case SUCCEEDED:
            commitment = null;
            if (goal.goalType() == Goal.MAINTAIN) {
            	state = NEW;
            } else {
                return false;
            }
		}
		
		return true;
	}

	private Bindings getBindingForPostCondition(IFormula post) {
		if (Predicate.class.isInstance(post)) {
			return Unifier.unifyFormulae((Predicate) post, (Predicate) goal.formula());
		} else if (AND.class.isInstance(post)) {
			AND and = (AND) post;
			for (IFormula part : and.asList()) {
				Bindings bindings = getBindingForPostCondition(part);
				if (bindings != null) return bindings;
			}
		}
        return null;
    }
	
	public List<Plan> options() {
		return options;
	}
	
	public Commitment commitment() {
		return commitment;
	}

	public Plan current() {
		return current;
	}
}
