package com.agentfactory.afapl2.compiler;


import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.afapl2.compiler.parser.Visitor;
import com.agentfactory.afapl2.compiler.parser.node.ActionNode;
import com.agentfactory.afapl2.compiler.parser.node.ActivityNode;
import com.agentfactory.afapl2.compiler.parser.node.CommitNode;
import com.agentfactory.afapl2.compiler.parser.node.FOSNode;
import com.agentfactory.afapl2.compiler.parser.node.ImpliesNode;
import com.agentfactory.afapl2.compiler.parser.node.ModalNode;
import com.agentfactory.afapl2.compiler.parser.node.Node;
import com.agentfactory.afapl2.compiler.parser.node.OntologyNode;
import com.agentfactory.afapl2.compiler.parser.node.PlanNode;
import com.agentfactory.afapl2.compiler.parser.node.ProgramNode;
import com.agentfactory.afapl2.compiler.parser.node.RoleNode;
import com.agentfactory.afapl2.compiler.parser.node.TriggerNode;
import com.agentfactory.afapl2.compiler.typeCheck.BasicTypeChecker;

/**
 * In this phase, the compiler checks that all activities used in plans and commitments
 * have been specified in the AFAPL code.  It also checks that all beliefs are specified
 * in the associated ontologies.
 *
 * @author Rem Collier
 */
public class Phase3Visitor extends Visitor {
    private List<PlanNode> plans;
    private List<ActionNode> actions;
    private BasicTypeChecker basicTypeChecker;
    private List<OntologyNode> ontologies;
    
    public Phase3Visitor(List<AFAPL2SyntaxError> errors, List<AFAPL2SyntaxError> warnings, Phase2Visitor p2Visitor) {
        super(errors,warnings);
        plans = p2Visitor.getPlans();
        actions = p2Visitor.getActions();
        ontologies = p2Visitor.getOntologies();
        basicTypeChecker = new BasicTypeChecker();
    }

    @Override
	public void visit(Node node) {
        if (node instanceof ProgramNode) {
            visit((ProgramNode) node);
        } else if (node instanceof RoleNode) {
            visit((RoleNode) node);
        } else if (node instanceof ImpliesNode) {
            visit((ImpliesNode) node);
        } else if (node instanceof CommitNode) {
            visit((CommitNode) node);
        } else if (node instanceof PlanNode) {
            visit((PlanNode) node);
        } else if (node instanceof TriggerNode) {
            visit((TriggerNode) node);
        } else {
            throw new RuntimeException("Attempted to visit Node: " + node + " of class: " + node.getClass().toString());
        }
    }
    
    private void visit(ProgramNode node) {
        for (ImpliesNode _node:node.getCommitmentRules()) {
            _node.accept(this);
        }
        for (PlanNode _node:node.getPlans()) {
            _node.accept(this);
        }
        for (ModalNode _node:node.getBeliefs()) {
       		_node.accept(new TypeCheckVisitor(ontologies, this.getWarnings(), this.getErrors()));
        }
        for (ImpliesNode _node:node.getBeliefRules()) {
            _node.accept(this);
        }
        for(CommitNode _node:node.getCommitments()) {
            _node.accept(this);
        }
    }
    
    private void visit(RoleNode node) {
    	for (ImpliesNode _node:node.getRules()) {
            _node.accept(this);
        }
        for (Node _node:node.getMentalState()) {
            _node.accept(this);
        }
        for (TriggerNode _node:node.getTriggers()) {
            _node.accept(this);
        }
    }
    
    private void visit(ImpliesNode node) {
        TypeCheckVisitor visitor = new TypeCheckVisitor(ontologies, this.getWarnings(), this.getErrors());
        node.getLHS().accept(visitor);
        
        if (node.getRHS() instanceof CommitNode) {
            CommitNode cNode = (CommitNode) node.getRHS();
            cNode.getCondition().accept(visitor);
            checkActivity(cNode.getActivity(), visitor.getTypeMap());
        } else {
            node.getRHS().accept(visitor);
        }
    }
        
    private void visit(CommitNode node) {
        TypeCheckVisitor visitor = new TypeCheckVisitor(ontologies, this.getWarnings(), this.getErrors());
        node.getCondition().accept(visitor);
        checkActivity(node.getActivity(), new HashMap<String, String>());
    }
    
    private void checkActivity(FOSNode node, Map<String, String> typeMap) {
        String functor = node.getFunctor();
        List<FOSNode> arguments = node.getArguments();
        if (functor.equals("PAR") || functor.equals("SEQ") ||
                functor.equals("OR") || functor.equals("XOR")) {
            for (int i=0; i<arguments.size();i++) {
                checkActivity(arguments.get(i), typeMap);
            }
        } else if (functor.equals("DO_WHEN") || functor.equals("FOREACH")) {
            // Need to check for variable introduction here...
            if (arguments.get(0).getFunctor().equals("BELIEF")) {
                TypeCheckVisitor typeCheck = new TypeCheckVisitor(ontologies, getWarnings(), getErrors(), typeMap);
                arguments.get(0).getArguments().get(0).accept(typeCheck);
            }

            checkActivity(arguments.get(1), typeMap);
        } else if (functor.equals("DO_UNTIL")) {
            checkActivity(arguments.get(1), typeMap);
        } else if (functor.equals("ATTEMPT")) {
            if (arguments.size() < 3) {
                addError(node.getSource(), node.getLine(), "Malformed Operator", "ATTEMPT operator takes three parameters, but only " + arguments.size() + " provided in " + node);
            }
            checkActivity(arguments.get(0), typeMap);
            checkActivity(arguments.get(1), typeMap);
            checkActivity(arguments.get(2), typeMap);
        } else if (functor.equals("AWAIT") || functor.equals("DELAY") ||
                   functor.equals("ADOPT") || functor.equals("RETRACT") ||
                   functor.equals("MAINTAIN")) {
        } else {
            validateActivity(node, typeMap);
        }
    }

    private void validateArguments(FOSNode node, ActivityNode activityNode, List<String> ontologies, Map<String, String> typeMap) {
        int j=0;
        for (FOSNode arg: node.getArguments()) {
            String type = activityNode.getType(j++);

            boolean found = false;

            if (arg.isVariable()) {
                if (typeMap.containsKey(arg.getFunctor())) {
                    found = type.equals(typeMap.get(arg.getFunctor()));
                } else {
                    found = true;
                }
            } else {
                int i=0;
                found = basicTypeChecker.validateTerm(type, arg);

                while (!found && (i<ontologies.size())) {
                    String ontologyName = ontologies.get(i++);
                    found = getOntology(ontologyName).validateArgument(type, arg);
                }
            }

           if (!found) {
                addError(node.getSource(), node.getLine(), "Syntax Error",
                		"Type Mismatch, expected: " + activityNode + " but got: " + node);
                return;
            }
        }
    }

    private void validateActivity(FOSNode node, Map<String, String> typeMap) {
        for (ActionNode actionNode: actions) {
            if (node.getFunctor().equals(actionNode.getIdentifier().getFunctor()) && 
                    (node.getArguments().size() == actionNode.getIdentifier().getVariables().size())) {
                // Need to check that each argument in the activity conforms 
                // with the type declaration given in the activity declaration
                validateArguments(node, actionNode.getIdentifier(), actionNode.getOntologies(), typeMap);
                
                // So, we have validated all of the arguments of the activity,
                // which means that we have validated the activity itself.
                return;
            }
        }

        for (PlanNode planNode: plans) {
            if (node.getFunctor().equals(planNode.getIdentifier().getFunctor()) &&
                    (node.getArguments().size() == planNode.getIdentifier().getVariables().size())) {

                // TODO: This needs to be replaced by an actual ontology for a plan...
                validateArguments(node, planNode.getIdentifier(), new LinkedList<String>(), typeMap);
                return;
            }
        }
        // At this point, all of the actions and plans have been checked
        // against the activity, but it has not been possible to validate
        // that activity.
        addError(node.getSource(), node.getLine(),"Unknown Activity",
        		"Could not find activity with identifier " + node);
    }
    
    private OntologyNode getOntology(String name) {
        for (OntologyNode node: ontologies) {
            if (node.getIdentifier().equals(name)) {
                return node;
            }
        }
        return null;
    }
    
    private void visit(PlanNode node) {
        // TODO: Redo this code as it is too low level - need to invoke this
        // code from the plan definition level

        Map<String, String> typeMap = new HashMap<String, String>();
        int i = 0;
        for (String var: node.getIdentifier().getVariables()) {
            String type = node.getIdentifier().getType(i++);
            typeMap.put(var, type);
        }
        checkActivity(node.getPlan(), typeMap);
    }
    
    private void visit(TriggerNode node) {
        node.getTrigger().accept( this );
    }
}
