/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package com.agentfactory.afapl2.compiler;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.agentfactory.afapl2.compiler.parser.Visitor;
import com.agentfactory.afapl2.compiler.parser.node.AlwaysNode;
import com.agentfactory.afapl2.compiler.parser.node.AndNode;
import com.agentfactory.afapl2.compiler.parser.node.BeliefNode;
import com.agentfactory.afapl2.compiler.parser.node.EqualNode;
import com.agentfactory.afapl2.compiler.parser.node.FOSNode;
import com.agentfactory.afapl2.compiler.parser.node.FunctionNode;
import com.agentfactory.afapl2.compiler.parser.node.GoalNode;
import com.agentfactory.afapl2.compiler.parser.node.GreaterThanNode;
import com.agentfactory.afapl2.compiler.parser.node.InRangeNode;
import com.agentfactory.afapl2.compiler.parser.node.LessThanNode;
import com.agentfactory.afapl2.compiler.parser.node.NextNode;
import com.agentfactory.afapl2.compiler.parser.node.Node;
import com.agentfactory.afapl2.compiler.parser.node.NotNode;
import com.agentfactory.afapl2.compiler.parser.node.OntologyNode;
import com.agentfactory.afapl2.compiler.parser.node.PredicateNode;
import com.agentfactory.afapl2.compiler.parser.node.UntilNode;

/**
 *
 * @author remcollier
 */
public class TypeCheckVisitor extends Visitor {
    private List<OntologyNode> ontologies;
    private Map<String, String> typeMap;
    
    public TypeCheckVisitor(List<OntologyNode> ontologies, List<AFAPL2SyntaxError> warnings, List<AFAPL2SyntaxError> errors) {
        this(ontologies, warnings, errors, new HashMap<String,String>());
    }
    
    public TypeCheckVisitor(List<OntologyNode> ontologies, List<AFAPL2SyntaxError> warnings,
    		List<AFAPL2SyntaxError> errors, Map<String, String> typeMap) {
        super(errors, warnings);
        
        this.ontologies = ontologies;
        this.typeMap = typeMap;
    }
    
    public Map<String, String> getTypeMap() {
        return typeMap;
    }
    
    @Override
	public void visit(Node node) {
        if (FOSNode.class.isInstance(node)) {
            visit((FOSNode) node);
        } else if (AndNode.class.isInstance(node)) {
            visit((AndNode) node);
        }  else if (NotNode.class.isInstance(node)) {
            visit((NotNode) node);
        }  else if (BeliefNode.class.isInstance(node)) {
            visit((BeliefNode) node);
        }  else if (GoalNode.class.isInstance(node)) {
            visit((GoalNode) node);
        }  else if (AlwaysNode.class.isInstance(node)) {
            visit((AlwaysNode) node);
        }  else if (NextNode.class.isInstance(node)) {
            visit((NextNode) node);
        }  else if (UntilNode.class.isInstance(node)) {
            visit((UntilNode) node);
        } else if (node instanceof LessThanNode) {
        } else if (node instanceof GreaterThanNode) {
        } else if (node instanceof EqualNode) {
        } else if (node instanceof InRangeNode) {
        } else {
            System.out.println("Unknown: " + node);
        }
    }
    
    private void visit(AlwaysNode node) {
        node.getContents().accept(this);
    }
    
    private void visit(NextNode node) {
        node.getContents().accept(this);
    }
    
    private void visit(NotNode node) {
        node.getContents().accept(this);
    }
    
    private void visit(UntilNode node) {
        node.getTerm().accept(this);
        node.getCondition().accept(this);
    }
    
    private void visit(BeliefNode node) {
        node.getContents().accept(this);
    }
    
    private void visit(GoalNode node) {
        node.getContents().accept(this);
    }
    
    private void visit(AndNode node) {
        for ( Node child : node.getChildren() ) {
            child.accept(this);
        }
    }
    
    private void visit(FOSNode node) {
        if (node.isTrueOrFalse()) return;

        // Search the ontologies looking for a matching predicaate
        for (OntologyNode ontologyNode: ontologies) {
            for (PredicateNode predicateNode: ontologyNode.getPredicates()) {

                if (predicateNode.getFunctor().equals(node.getFunctor()) &&
                        predicateNode.numArguments() == node.getArguments().size()) {

                    // We have found a matching predicate, so now we need to
                    // validate the arguments of our term against the predicate
                    // definition...
                    int i=0;
                    for (FOSNode arg: node.getArguments()) {
                        String type = predicateNode.getType(i++);
                        if (arg.isVariable()) {
                            // Here, the argument is a variable, so we need to
                            // check whether this is the first appearance of the
                            // variable, or whether it was previously specified.
                            if (typeMap.containsKey(arg.getFunctor())) {
                                // Variable exists, so check that types match
                                if (!typeMap.get(arg.getFunctor()).equals(type)) {
                                    addError(node.getSource(), node.getLine(),"Type Mismatch",
                                        "the variable: " + arg +  " in: " + node +
                                        "' has type: " + typeMap.get(arg.getFunctor()) +
                                        " but requires type: " + type + "\n");
                                }
                            } else {
                                // Variable does not exist, so add it to the
                                // type map.
                                typeMap.put(arg.getFunctor(), type);
                            }
                        } else if (arg.isComposite()) {
                            validateComposite(type, node, ontologyNode, arg);
                        } else {
                            if (!ontologyNode.validateArgument(type, arg)) {
                                addError(node.getSource(), node.getLine(), "Type Mismatch",
                                		"the argument: " + arg + " in: " + node +
                                		"' does not match its type: " + type + "\n");
                            }
                        }
                    
                    }
                    return;
                }
            }
        }
        
        // Predicate is not defined, so extract the variables, set them as type
        // "none", and output a warning.
        VariableExtractor extractor = new VariableExtractor();
        node.accept(extractor);
        for (String var: extractor.getVariables()) {
            if (!typeMap.containsKey(var)) {
                typeMap.put(var, "none");
            }
        }

        addWarning(node.getSource(), node.getLine(),"Undefined Content Language Term",
        		"a predicate matching '" + node + "' could not be found in any ontology");
    }

    private boolean isPrimitiveType(String type) {
        return type.equals("short") || type.equals("int") || type.equals("long") ||
               type.equals("float") || type.equals("double") || type.equals("boolean") ||
               type.equals("string") || type.equals("char");
    }

    private void validateComposite(String type, FOSNode node, FOSNode arg) {
        if (isPrimitiveType(type)) {
            // We have a primitive type, so throw a warning
            addError(node.getSource(), node.getLine(), "Type Mismatch",
            		"the argument: " + arg + " in: " + node +
            		"' can not be matched against the primitive type: " + type);
        } else if (type.equals("none") ||type.equals(arg.getFunctor())) {
            // We have a functional type, so validate the argument
            for (OntologyNode ontologyNode:ontologies) {
                for (FunctionNode functionNode: ontologyNode.getFunctions()) {
                    if (functionNode.getFunctor().equals(arg.getFunctor()) &&
                       (functionNode.getVariables().size() == arg.getArguments().size())) {

                        validateFunction(node, ontologyNode, functionNode, arg);
                        return;
                    }
                }
            }

            // Whoops, we didn't find anything...
            // So, we need extract any internal variables and set them up with
            // type "none"...
            VariableExtractor extractor = new VariableExtractor();
            arg.accept(extractor);
            for (String var: extractor.getVariables()) {
                if (!typeMap.containsKey(var)) {
                    typeMap.put(var, "none");
                }
            }
        } else {
            // We have no idea what we have and so, lets throw
            // a generic error.
            
            // TODO: Fix this
            addError(node.getSource(), node.getLine(),"Type Mismatch",
            		"ERROR: the argument: " + arg + " in: " + node +
            		"' does not match its type: " + type + "\n");
        }
    }

    private void validateComposite(String type, FOSNode node, OntologyNode ontologyNode, FOSNode arg) {
        // Here the argument is a composite, meaning that we must match it
        // against one of the functions in the ontology. It cannot be a primitive
        // type, however, if the type is "none", then it can be matched against
        // anything...
        if (isPrimitiveType(type)) {
            // We have a primitive type, so throw a warning
            addError(node.getSource(), node.getLine(), "Type Mismatch",
            		"the argument: " + arg + " in: " + node +
            		"' can not be matched against the primitive type: " + type);
        } else if (type.equals("none")) {
            // We have the none type, so we can match against
            // anything

            validateComposite(type, node, arg);
        } else if (type.equals(arg.getFunctor())) {
            // We have a functional type, so validate the
            // argument
            for (FunctionNode functionNode: ontologyNode.getFunctions()) {
                if (functionNode.getFunctor().equals(arg.getFunctor()) &&
                   (functionNode.getVariables().size() == arg.getArguments().size())) {

                    validateFunction(node, ontologyNode, functionNode, arg);
                    return;
                }
            }

        } else {
            // We have no idea what we have and so, lets throw
            // a generic error.
            // TODO: Fix this
            addError(node.getSource(), node.getLine(), "Type Mismatch", 
            		"the argument: " + arg + " in: " + node +
            		"' does not match its type: " + type);
        }
    }

    private void validateFunction(FOSNode node, OntologyNode ontologyNode, FunctionNode functionNode, FOSNode arg) {
        // We have found the matching function term in the ontology,
        // so start matching it against the argument.
        int i = 0;
        for (FOSNode argument:arg.getArguments()) {
            String argType = functionNode.getTypes().get(i++);

            if (arg.isVariable()) {
                if (typeMap.containsKey(arg.getFunctor())) {
                    if (!typeMap.get(arg.getFunctor()).equals(argType)) {
                        addError(node.getSource(), node.getLine(), "Type Mismatch",
                        		"the variable: " + argument + " in: " + node +
                        		"' has type: " + typeMap.get(arg.getFunctor()) +
                            " but requires type: " + functionNode.getTypes().get(i));
                    }
                } else {
                    typeMap.put(argument.getFunctor(), argType);
                }
            } else if (arg.isComposite()) {
                validateComposite(argType, node, ontologyNode, argument);
            }
        }

    }
}
