package com.agentfactory.afapl2.compiler.parser;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.afapl2.compiler.parser.reader.Reader;
import com.agentfactory.afapl2.compiler.parser.token.Token;
/**
 * This class generates a set of lexical tokens that represent the
 * AF-APL that is being compiled.
 *
 * @author  Rem Collier
 */
public class LexicalAnalyser {
    private Reader reader;
    private int line;
    private int character;
    private List<Token> buffer;
    private StringBuffer tempBuffer;
    
    /** Creates a new instance of LexicalAnalyser */
    public LexicalAnalyser(Reader reader) {
        this.reader = reader;
        line = 1;
        character = 0;
        buffer = new LinkedList<Token>();
    }
    
    private void incrementLine() {
        line++;
        character = 0;
    }
    
    private char getCharacter() {
        character++;
        char ch = reader.nextCharacter();
        tempBuffer.append(ch);
        return ch;
    }
    
    private void pushBackCharacter(char character) {
        reader.pushBack(character);
    }
    
    public Token getNextToken() throws LexicalAnalyserException {
        if (buffer.size() > 0) {
            return buffer.remove(0);
        }
        
        tempBuffer = new StringBuffer();
        
        boolean finished = false;
        boolean error = false;
        boolean stop;
        
        Token token = null;
        StringBuffer buf = null;
        
        char ch;
        while (!finished) {
            ch = getCharacter();
            
            // DEAL WITH SPECIAL CASES:
            if (ch == '\n') {
                // carriage return!
                incrementLine();
            } else if ((ch == '\t') || (ch == '\r') || (ch == ' ')) {
                // do nothing!
            } else if (ch == '/') {
                // possible comment!
                ch = getCharacter();
                if (ch == '/') {
                    while ((ch = getCharacter()) != '\n');
                    incrementLine();
                } else if (ch == '*') {
                    stop = false;
                    while (!stop) {
                        ch = getCharacter();
                        if (ch == '*') {
                            ch = getCharacter();
                            if (ch == '/') {
                                stop = true;
                            } else {
                                reader.pushBack(ch);
                            }
                        }

                        if (ch == '\n') {
                            incrementLine();
                        }
                    }
                }
            } else if (ch == '\"') {
                buf = new StringBuffer();
                buf.append(ch);

                while ((ch = getCharacter()) != '\"') {
                    if (ch == '\\') {
                        // this is an escape character, so just add the next
                        // character as is "must" be a valid character.
                        buf.append(ch);
                        ch = getCharacter();
                    }
                    buf.append(ch);
                }
                buf.append(ch);
                
                token = Token.createToken(buf.toString(), line, reader);
                finished = true;
            } else if (ch == '\0') {
                finished = true;
            } else if (alphaNumeric(ch)) {
                buf = new StringBuffer();
                buf.append(ch);
                
                ch = getCharacter();
                while (alphaNumeric(ch)) {
                    buf.append(ch);
                    ch = getCharacter();
                }

                reader.pushBack(ch);
                
                // Check for type!
                token = Token.createToken(buf.toString(), line, reader);
                finished = true;
            } else if (ch == '?') {
                buf = new StringBuffer();
                buf.append(ch);
                
                ch = getCharacter();
                while (alphaNumeric(ch)) {
                    buf.append(ch);
                    ch = getCharacter();
                }

                reader.pushBack(ch);
                token = Token.createToken(buf.toString(), line, reader);
                finished = true;
            } else if (ch == ';') {
                token = Token.createToken(";", line, reader);
                finished = true;
            } else if (ch == ',') {
                token = Token.createToken(",", line, reader);
                finished = true;
            } else if (ch == '(') {
                token = Token.createToken("(",  line, reader);
                finished = true;
            } else if (ch == ')') {
                token = Token.createToken(")",  line, reader);
                finished = true;
            } else if (ch == '{') {
                token = Token.createToken("{",  line, reader);
                finished = true;
            } else if (ch == '}') {
                token = Token.createToken("}",  line, reader);
                finished = true;
            } else if (ch == '&') {
                token = Token.createToken("&",  line, reader);
                finished = true;
            } else if (ch == '!') {
                token = Token.createToken("!",  line, reader);
                finished = true;
            } else if (ch == '^') {
                token = Token.createToken("^",  line, reader);
                finished = true;
            } else if (ch == ':') {
                token = Token.createToken(":", line, reader);
                finished = true;
            } else if (ch == '=') {
                ch = getCharacter();
                if (ch == '>') {
                    token = Token.createToken("=>",  line, reader);
                    finished = true;
                } else {
                    token = Token.createToken("=", line, reader);
                    this.pushBackCharacter(ch);
                    finished = true;
                }
            } else {
                error = true;
            }
            
            if (error) {
                throw new LexicalAnalyserException(this.getSource(), line, "Parse Error near: " + tempBuffer.toString());
            }
        }
        
        return token;
    }
    
    public void back(Token token) {
        buffer.add(0, token);
    }
    
    /**
     * TODO - RR - Rem, this is too restrictive; what about a hyphen for example?
     */
    private boolean alphaNumeric(char ch) {
        if ((ch >= 'A') && (ch <= 'Z')) {
            return true;
        }
        
        if ((ch >= 'a') && (ch <= 'z')) {
            return true;
        }
        
        if ((ch >= '0') && (ch <= '9')) {
            return true;
        }
        
        if ((ch == '_') || (ch == '-') || (ch == '.') || (ch == '+') ||
            (ch == '@') || (ch == '/') || (ch == '*') || (ch == '>') ||
            (ch == '<')) {
            return true;
        }
        
        return false;
    }
    
    public String getSource() {
        return reader.toString();
    }
}

