package com.agentfactory.afapl2.interpreter;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.afapl2.interpreter.control.GoalManager;
import com.agentfactory.afapl2.interpreter.control.PlanLibrary;
import com.agentfactory.afapl2.interpreter.control.RoleLibrary;
import com.agentfactory.afapl2.interpreter.control.cms.CommitmentManager;
import com.agentfactory.afapl2.interpreter.mentalState.Action;
import com.agentfactory.afapl2.interpreter.mentalState.Commitment;
import com.agentfactory.logic.MentalAgent;
import com.agentfactory.logic.agent.Module;
import com.agentfactory.logic.reasoner.SubstitutionSet;

/**
 * This is the implementation of the Agent Factory Agent class for the AF-APL
 * programming language.
 *
 * The AFAPLAgent class creates and holds all control elements used to
 * run the logical AFAPL agent. This does not includes meta constructs such as the
 * agent's name, or the services it uses - these are managed by the super Agent class.
 *
 * @author <a href="mailto:robertr@tzi.de">Robert Ross </a>
 */
public abstract class AbstractAFAPL2Agent extends MentalAgent {

    protected IGoalManager goalManager;
    protected ICommitmentManager commitmentManager;
    protected Map<String, Action> actions;
    protected Map<String, Module> modules;
    protected RoleLibrary roleLibrary;
    protected PlanLibrary planLibrary;
    private List<ScheduleItem> schedule;
    private List<CompletedActivity> completedItems;

    /**
     * Construct an AF-APL interpreter. This constructor is used by the
     * AgentPlatform class via a factory method.
     *
     * This constructor firstly calls the super class constructor, passing its
     * name, type and a reference to the AgentPlatform object. The agent design
     * is then read in from file and used to instantiate the various agent
     * containers.
     *
     * @param inName - A string object. The given name of the agent.
     * @param inType - A String object. The agent type -- i.e. the type of agent that
     * is to be substantiated e.g. SuperDF.agt'
     * @param inAgentPlatform - An AgentPlatform object. A reference to the AF platfrom
     * on which this agent is being created.
     * @throws AgentConfigurationException -- Throw if there were any major problems
     * creating an agent. Naturally, if this exception is thrown during construction,
     * the agent is not usable.
     */
    public AbstractAFAPL2Agent( String name ) {
        super( name );

        Logger.getLogger( name ).setLevel( Level.SEVERE );

        commitmentManager = new CommitmentManager( this );
        goalManager = new GoalManager( this );
        queryEngine.addSource( goalManager );
        roleLibrary = new RoleLibrary( this );
        planLibrary = new PlanLibrary();
        modules = new HashMap<String, Module>();
        actions = new HashMap<String, Action>();
        schedule = new ArrayList<ScheduleItem>();
        completedItems = new ArrayList<CompletedActivity>();
    }

    /**
     * Iterate through any of the arguments supplied in the platform configuration
     * file, and set class values accordingly.
     * Arguments are expected as PARAMETER=value strings.
     */
    public void processArgs( Properties arguments ) {
        String argument = arguments.getProperty( "TIMESLICE" );
        if ( argument != null ) {
            setTimeSlice( Integer.parseInt( argument ) );
        }
    }

    public synchronized void execute() {
        Logger.getLogger( name ).log( Level.INFO, "In Execute" );
        this.retrieveNewMessages();

        // Update the beliefs (perception + temporal update)
        this.beliefManager.update();
        Logger.getLogger( name ).log( Level.INFO, "Updated Beliefs" );

        // clear the inbox
        getOutbox().clear();
        Logger.getLogger( name ).log( Level.INFO, "Cleared Out Box" );

        // manage goals
        goalManager.manage();
        Logger.getLogger( name ).log( Level.INFO, "Managed Goals" );

        // manage roles
        roleLibrary.enactRoles();
        Logger.getLogger( name ).log( Level.INFO, "Enacted Roles" );

        // manage commitments
        commitmentManager.adoptCommitments();
        Logger.getLogger( name ).log( Level.INFO, "Adopted Commitments" );
        commitmentManager.manageCommitments();
        Logger.getLogger( name ).log( Level.INFO, "Managed Commitments" );

        // perform actions
        long duration = 0;
        ScheduleItem item = null;
        while ( !schedule.isEmpty() && !isEndOfTimeSlice() ) {
            item = schedule.remove( 0 );
            Logger.getLogger( name ).log( Level.INFO, "Performing: " + item.action );
            try {
                duration = System.currentTimeMillis();
                boolean result = item.action.getActuator().act( item.commitment.getActivity() );
                duration = System.currentTimeMillis() - duration;
                completedItems.add( new CompletedActivity( item.commitment.getActivity(), duration, result ) );
                if ( result ) {
                    item.commitment.update( Commitment.EVENT_SUCCEEDED );
                }
                else {
                    item.commitment.update( Commitment.EVENT_FAILED );
                }
            }
            catch ( RuntimeException re ) {
                re.printStackTrace();
                item.commitment.update( Commitment.EVENT_FAILED );
            }
            Logger.getLogger( name ).log( Level.INFO, "Completed: " + item.action );
        }
        Logger.getLogger( name ).log( Level.INFO, "Performed Actions" );

        // Finish by clearing the inbox...
        this.endOfIteration();
    }

    /**
     * This method is called from the commitment manager and is used to
     * schedule actions that need to be performed.
     * 
     * @param commitment the commitment whose action is being scheduled...
     */
    public void scheduleAction( Commitment commitment, Action action ) {
        Logger.getLogger( name ).log( Level.FINE, "[" + getName() + "]: Scheduling: " + commitment.getActivity().toString() );
        schedule.add( new ScheduleItem( commitment, action ) );
    }

    public List<String> getActionSchedule() {
        List<String> list = new ArrayList<String>();
        for ( ScheduleItem item : schedule ) {
            list.add( item.commitment.getActivity().toString() );
        }
        return list;
    }

    private class ScheduleItem {

        Commitment commitment;
        Action action;

        public ScheduleItem( Commitment commitment, Action action ) {
            this.commitment = commitment;
            this.action = action;
        }
    }

    /***************************************************************************
     * ACCESSOR METHODS
     ***************************************************************************/
    public IGoalManager getGoalManager() {
        return goalManager;
    }

    public ICommitmentManager getCommitmentManager() {
        return commitmentManager;
    }

    public Map<String, Action> getActions() {
        return actions;
    }

    public PlanLibrary getPlanLibrary() {
        return planLibrary;
    }

    public RoleLibrary getRoleLibrary() {
        return roleLibrary;
    }

    public String getType() {
        return "AFAPL2";
    }

    @Override
    public synchronized Module getModuleByName( String name ) {
        return modules.get( name );
    }

    @Override
    public synchronized Module getModuleByClass( String className ) {
        Iterator it = modules.values().iterator();
        Module module = null;
        boolean found = false;
        while ( !found && it.hasNext() ) {
            module = (Module) it.next();
            found = module.getClass().getName().equals( className );
        }

        if ( !found ) {
            return null;
        }

        return module;
    }

    /**
     * True get-by-class method to find a module based on a {@code Class} object.
     *
     * @param lookup The class of the module to be found
     * @return the first module matching the given class
     */
    public synchronized <T extends Module> T getModuleByClass( Class<T> lookup ) {
        for ( Module m : modules.values() ) {
            if ( lookup.isInstance( m ) ) {
                return (T) m;
            }
        }
        return null;
    }

    public synchronized Module[] getModulesByClass( String className ) {
        List<Module> list = new LinkedList<Module>();
        for ( Module module : modules.values() ) {
            if ( module.getClass().getName().equals( className ) ) {
                list.add( module );
            }
        }

        return (Module[]) list.toArray();
    }

    /**
     * True get-by-class method to find modules based on a {@code Class} object.
     *
     * @param lookup The class of the modules to be found
     * @return all modules matching the given class
     */
    public synchronized Module[] getModulesByClass( Class lookup ) {
        List<Module> toReturn = new LinkedList<Module>();
        for ( Module m : modules.values() ) {
            if ( m.getClass().equals( lookup ) ) {
                toReturn.add( m );
            }
        }
        return (Module[]) toReturn.toArray();
    }

    public synchronized Module[] getAllModules() {
        return (Module[]) modules.values().toArray();
    }

    /***************************************************************************
     * MENTAL STATE QUERY METHODS
     ***************************************************************************/
    public boolean query( Object goal ) {
        return queryEngine.query( goal );
    }

    public List<SubstitutionSet> queryAll( Object goal ) {
        if ( queryEngine.queryAll( goal ) ) {
            return queryEngine.getSubstitutionSets();
        }
        return null;
    }

    public List<CompletedActivity> getCompletedItems() {
        List<CompletedActivity> list = completedItems;
        completedItems = new ArrayList<CompletedActivity>();
        return list;
    }
}
// vim:sw=4:

