/**
 * Title:       CommitmentManager.java
 * Copyright:   Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved. 
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */
package com.agentfactory.afapl2.interpreter.control.cms;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.agentfactory.afapl2.interpreter.AbstractAFAPL2Agent;
import com.agentfactory.afapl2.interpreter.ICommitmentManager;
import com.agentfactory.afapl2.interpreter.mentalState.Commitment;
import com.agentfactory.afapl2.interpreter.mentalState.CommitmentRule;
import com.agentfactory.afapl2.interpreter.mentalState.Role;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.reasoner.SubstitutionSet;

/**
 * Superclass for all commitment mangers. This class will eventually contain
 * functionality common to all commitment managers.
 * 
 * @author <a href="mailto:robertr@tzi.de">Robert Ross</a>
 * @author Rem Collier
 */
public class CommitmentManager implements ICommitmentManager {
    private static final Logger logger = Logger.getLogger( CommitmentManager.class.getName() );

    static {
        logger.setLevel( Level.WARNING );
    }

    private AbstractAFAPL2Agent agent;
    private List<CommitmentRule> commitmentRules;
    private List<Commitment> commitments; // TODO(claudiagrigore): make it a set

    /**
     * Construct a basic CommitmentManager
     */
    public CommitmentManager( AbstractAFAPL2Agent agent ) {
        this.agent = agent;
        commitmentRules = new LinkedList<CommitmentRule>();
        commitments = new LinkedList<Commitment>();
    }

    protected void setAgent( AbstractAFAPL2Agent agent ) {
        this.agent = agent;
    }

    /**
     * Manage the agents commitments. This method should process commitments
     * only so long as it has time to process them.
     * 
     * @param startTime
     *            the start time of the current time slice.
     */
    public void manageCommitments() {
        int index = 0;
        boolean remove = false;
        setEndOfTimeSlice( false );
        Commitment nextCommitment = null;
        while ( index < commitments.size() && !endOfTimeSliceReached() ) {
            nextCommitment = commitments.get( index );

            if ( nextCommitment == null ) {
                throw new RuntimeException( "error, could not get the next commmitment" );
            }

            // Process the next commitment, generating a set of new commitments
            // to be added into the commitments for the next manageCommitments()
            // call
            synchronized ( nextCommitment ) {
                if ( processCommitment( nextCommitment ) ) {
                    index++;
                }
                else {
                    nextCommitment.completed();
                    commitments.remove( index );
                }
            }
        }
    }

    private boolean endOfTimeSlice;

    private synchronized boolean endOfTimeSliceReached() {
        return endOfTimeSlice;
    }

    public synchronized void setEndOfTimeSlice( boolean value ) {
        endOfTimeSlice = value;
    }

    public boolean hasCommitment( Commitment commitment ) {
        return commitments.contains( commitment );
    }

    public List<Commitment> getCommitments() {
        return commitments;
    }

    public List<CommitmentRule> getCommitmentRules() {
        List<CommitmentRule> ruleSet = new ArrayList<CommitmentRule>();
        ruleSet.addAll( commitmentRules );
        for ( Role role : agent.getRoleLibrary().getInstances() ) {
            ruleSet.addAll( role.getCommitmentRules() );
        }
        return ruleSet;
    }

    private void addCommitment( Commitment c ) {
        c.applyInPlace( basicSubstitution() ); // don't create a new Commitment
                                               // object
        c.evaluate();
        commitments.add( c );
    }

    public void addInitialCommitment( Commitment c ) {
        addCommitment( c );
    }

    /**
     * Take the mental state of the agent i.e. the commitment rules and beliefs
     * and generate the set of new primary commitments which should be created
     * as a result of this metnal state.
     * 
     */
    public synchronized void adoptCommitments() {
        Commitment commitment = null;

        // Generate the "Active" Commitment Rule Set
        // This is generic rules + enacted Role rules
        for ( CommitmentRule commitmentRule : getCommitmentRules() ) {
            logger.log( Level.INFO, "Checking Rule: " + commitmentRule );

            // Check whether the LHS of the commitment manager matches anything
            // in the ages current mental state.
            List<SubstitutionSet> bindings = agent.queryAll( commitmentRule.getSituation() );
            logger.log( Level.INFO, "Bindings: " + bindings );

            // if there were any matches ...
            if ( bindings != null ) {
                // We have matched the left-hand side of the commitment
                // rule to the agents beliefs, so start processing
                // each belief (i.e. create the resultant commitments)

                if ( bindings.isEmpty() ) {
                    System.err.println( "XXX: match with no bindings!" );
                }

                // If there were bindings to be matched....
                for ( SubstitutionSet set : bindings ) {
                    commitment = ( (Commitment) commitmentRule.getCommitment().clone() ).apply( set );
                    addCommitment( commitment );

                    // Keep track of which rule generated which commitment
                    commitment.setCommitmentRule( commitmentRule );
                    commitmentRule.addCommitment( commitment );
                }
            }
        }
    }

    public void addCommitmentRule( CommitmentRule commitmentRule ) {
        commitmentRules.add( commitmentRule );
    }

    private static final FOS self = new FOS( "?self" );
    private static final FOS now = new FOS( "?now" );

    /* Give values to "?self" and to "?now". */
    private SubstitutionSet basicSubstitution() {
        SubstitutionSet ss = new SubstitutionSet();
        ss.bind( self, new FOS( agent.getName() ) );
        ss.bind( now, new FOS( "after(0)" ) );
        return ss;
    }

    private boolean processCommitment( Commitment commitment ) {
        ActivityHandler handler = null;

        if ( commitment.getHandler() == null ) {
            // Step 1: Lazy initialization of handler
            if ( commitment.getState() == Commitment.STATE_NEW ) {
                handler = ActivityHandler.getInstance( commitment, agent );
            }

            // Here we will either have not new state / handler not assigned /
            // or a valid handler...
            if ( handler == null ) {
                Commitment parent = commitment.getParent();
                if ( parent != null ) {
                    parent.update( Commitment.EVENT_CHILD_FAILED );
                }
                return false;
            }
        }
        else {
            handler = commitment.getHandler();
        }

        return handler.handle();
    }
}
// vim:sw=4:
