/**
 * Title:       Commitment.java
 * Copyright:   Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved.
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */
package com.agentfactory.afapl2.interpreter.mentalState;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;

import com.agentfactory.afapl2.interpreter.control.cms.ActivityHandler;
import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.lang.TerAND;
import com.agentfactory.logic.reasoner.SubstitutionSet;
import java.util.List;

/**
 * Class used to contain a full commitment. A commitment is a four tuple
 * String AFAPLAgent
 * Calendar startTime
 * BeliefString maintanence condition
 * Activity activity (plan / action commited to)
 *
 * TODO - RR - Refactor the commitment state machine introducing the INACTIVE
 * and ACTIVE states to replace the WAITING state.
 *
 * @author <a href="mailto:robertr@tzi.de">Robert Ross</a>
 *
 */
public class Commitment {
    // States

    public static final int STATE_NEW = 0;
    public static final int STATE_ACTIVE = 1;
    public static final int STATE_WAITING = 2;
    public static final int STATE_FAILED = 3;
    public static final int STATE_SUCCEEDED = 4;
    public static final int STATE_REDUNDANT = 5;
    // Events
    public static final int EVENT_ACTIVATED = 10;
    public static final int EVENT_SUCCEEDED = 11;
    public static final int EVENT_FAILED = 12;
    public static final int EVENT_CHILD_SUCCEEDED = 13;
    public static final int EVENT_CHILD_FAILED = 14;
    public static final int EVENT_PARENT_FAILED = 15;
    public static final int EVENT_REDUNDANT = 16;
    // Data members
    private Commitment parent;
    private List<Commitment> children;
    private FOS to;
    private FOS startTime;
    private TerAND maintenanceCondition;
    private FOS activity;
    private int state;
    private ActivityHandler handler;
    private CommitmentRule commitmentRule;

    public Commitment( FOS _to, FOS _startTime, TerAND _maintenanceCondition, FOS _activity ) {
        this( _to, _startTime, _maintenanceCondition, _activity, null );
    }

    /** Creates a new instance of COMMIT */
    public Commitment( FOS _to, FOS _startTime, TerAND _maintenanceCondition,
                       FOS _activity, Commitment _parent ) {
        to = _to;
        maintenanceCondition = _maintenanceCondition;
        state = Commitment.STATE_NEW;
        startTime = _startTime;
        activity = _activity;

        // Handle Commitment Structure Hierarchy
        parent = _parent;
        children = new ArrayList<Commitment>();
        if ( parent != null ) {
            parent.addChild( this );
        }
    }
    private static final SimpleDateFormat timeFormatter =
            new SimpleDateFormat( "HH.mm" );

    private static FOS evalAfter( int delay ) {
        long now = System.currentTimeMillis();
        Calendar time = Calendar.getInstance();
        time.add( Calendar.MILLISECOND, delay );
        return new FOS( timeFormatter.format( time.getTime() ) );
    }

    private static FOS evaluate( FOS f ) {
        assert f != null;
        if ( f.getFunctor().equals( "after" ) ) {
            // Handle "after", the only interpreted function (for now).
            assert f.numArguments() == 1 : "The compiler should reject such horrible code.";
            return evalAfter( Integer.valueOf( f.argAt( 0 ).getFunctor() ) );
        }
        else {
            // Recurse on non-interpreted functions.
            boolean changed = false;
            ArrayList<FOS> na = new ArrayList<FOS>();
            for ( int i = 0; i < f.numArguments(); ++i ) {
                FOS a = evaluate( f.argAt( i ) );
                changed |= a != f.argAt( i );
                na.add( a );
            }
            return changed ? new FOS( f.getFunctor(), na ) : f;
        }
    }

    /** 
     * Evaluates interpreted functions such as |after|. Also, it evaluates the
     * special syntax '+30', which means "after 30 minutes" and may appear
     * only in |startTime|.
     */
    public void evaluate() {
        to = evaluate( to );
        startTime = evaluate( startTime );
        activity = evaluate( activity );

        TerAND mc = new TerAND();
        boolean changed = false;
        for ( FOS f : maintenanceCondition ) {
            FOS nf = evaluate( f );
            changed |= f != nf;
            mc.add( nf );
        }
        if ( changed ) {
            maintenanceCondition = mc;
        }

        // Handles the '+<minutes>' construct.
        String t = startTime.getFunctor();
        if ( t.matches( "\\+[0-9]+" ) ) {
            int m = Integer.valueOf( t.substring( 1 ) );
            assert m <= Integer.MAX_VALUE / 60 / 10000 :
                    "The compiler should complain about such big values.";
            startTime = evalAfter( m * 60 * 1000 );
        }
    }

    public FOS getTo() {
        return to;
    }

    public FOS getStartTime() {
        return startTime;
    }

    public void setStartTime( FOS startTime ) {
        this.startTime = startTime;
    }

    public TerAND getMaintenanceCondition() {
        return maintenanceCondition;
    }

    public int getState() {
        return state;
    }

    public FOS getActivity() {
        return activity;
    }

    public Commitment getParent() {
        return parent;
    }

    public void setHandler( ActivityHandler _handler ) {
        handler = _handler;
    }

    public ActivityHandler getHandler() {
        return handler;
    }

    public void addChild( Commitment commitment ) {
        if ( this.equals( commitment.getParent() ) ) {
            children.add( commitment );
        }
    }

    @Override
    public Object clone() {
        return new Commitment( (FOS) to.clone(), (FOS) startTime.clone(), (TerAND) maintenanceCondition.clone(),
                (FOS) activity.clone(), parent );
    }

    public Commitment apply( SubstitutionSet set ) {
        return new Commitment( to.apply( set ), startTime.apply( set ),
                maintenanceCondition.apply( set ),
                activity.apply( set ), parent );
    }

    /**
     * Apply a SubstitutionSet to this commitment without creating a new Commitment
     * @param set
     */
    public void applyInPlace( SubstitutionSet set ) {
        to.apply(  set );
        startTime.apply( set );
        maintenanceCondition.apply( set );
        activity.apply( set );
    }

    public synchronized void update( int event ) {
//        System.out.println("Commitment: " + this + " state: " + event + " handler: " + handler);
        if ( handler != null ) {
            // If handler is null, then the commitment has been completed by 
            // another process
            handler.handleEvent( event );
        }
    }

    public void setState( ActivityHandler _handler, int _state ) {
        if ( handler.equals( _handler ) ) {
            state = _state;
        }
    }

    public void completed() {
        children = null;
        parent = null;
        handler = null;
    }

    @Override
    public String toString() {
        return "COMMIT(" + to + "," + startTime + "," + maintenanceCondition + "," + activity + ")";
    }

    public synchronized boolean removeCompletedChildren() {
        Commitment child = null;
        int index = 0;
        while ( index < children.size() ) {
            child = (Commitment) children.get( index );
            if ( child.getState() == Commitment.STATE_SUCCEEDED ) {
                children.remove( index );
            }
            else {
                index++;
            }
        }

        return children.isEmpty();
    }

    public synchronized boolean removeFailedChildren() {
        Commitment child = null;
        int index = 0;
        while ( index < children.size() ) {
            child = (Commitment) children.get( index );
            if ( child.getState() == Commitment.STATE_FAILED ) {
                children.remove( index );
            }
            else {
                index++;
            }
        }
        return children.size() == 0;
    }

    public void notifyChildren( int event ) {
        Commitment child = null;
        int index = 0;
        while ( index < children.size() ) {
            child = (Commitment) children.get( index++ );
            child.update( event );
        }
    }

    public boolean hasChildren() {
        if ( children == null ) {
            return false;
        }
        return children.size() > 0;
    }

    public List<Commitment> getChildren() {
        return children;
    }

    public void setCommitmentRule( CommitmentRule rule ) {
        commitmentRule = rule;
    }

    public CommitmentRule getCommitmentRule() {
        return commitmentRule;
    }

    /**
     * Check equality of Commitments
     * @param o Another instance of {@code Commitment} to compare with
     * @return {@code true} if the agent name, start time, maintenance condition and activity are equal, {@code false} otherwise.
     */
    @Override
    public boolean equals( Object o ) {
        return super.equals( o ) || ( o != null && toString().equals( o.toString() ) );
    }
}
