package com.agentfactory.afapl2.interpreter.mentalState;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.logic.lang.FOS;

/**
 * Representation of GOALS in the system
 *
 * @author <a href="mailto:rem.collier@ucd.ie">Rem Collier</a>
 *
 */
public class Goal extends FOS {
    /**
     * Newly-created goal
     */
    public static final int STATE_NEW                                       = 0;
    
    /**
     * Ready for execution (the agent's activities have been analysed to 
     * find ones that should result in satisfaction of the goal)
     */
    public static final int STATE_READY                                     = 1;
    
    /**
     * The agent is in the process of attempting to satisfy this goal
     */
    public static final int STATE_ACTIVE                                    = 2;
    
    /**
     * The agent's attempt to satisfy the goal was not successful
     */
    public static final int STATE_FAILED                                    = 3;
    
    /**
     * The agent's attempt to satisfy the goal was successful
     */
    public static final int STATE_SUCCEEDED                                 = 4;
    
    private class Option {
        static final int STATE_NEW                                          = 0;
        static final int STATE_ACTIVE                                       = 1;
        
        FOS activity;
        int state;
        
        public Option(FOS activity) {
            this.activity = activity;
            state = STATE_NEW;
        }
    }
    
    private int state;
    private List<Option> options;
    private Commitment commitment;
    private boolean isToBeMantained;
    private FOS goalAsBelief;
    
    /**
     * Construct a belief directly from a string representation
     */
    public Goal(String inString) {
        super(inString);
        
        goalAsBelief = new FOS("BELIEF(" + argAt(0).toString() + ")");
        state = STATE_NEW;
        options = new LinkedList<Option>();
        
        // Make sure we meet the definition of a GOAL
        isToBeMantained = false;
        if( !getFunctor().equals("GOAL") ) {
            throw new IllegalArgumentException("That did not appear to be a Goal");
        }
    }
    
    public FOS getCorrespondingBelief() {
        return goalAsBelief;
    }
    
    public int getState() {
        return state;
    }
    
    public void setState(int _state) {
        state = _state;
    }
    
    public void addOption(FOS option) {
        options.add(new Option(option));
    }
    
    public List<FOS> getOptions() {
        List<FOS> list = new LinkedList<FOS>();
        for (Option option: options) {
            list.add(option.activity);
        }
        return list;
    }
    
    public FOS getNextOption() {
        for (Option option: options) {
            if (option.state == Option.STATE_NEW) {
                option.state = Option.STATE_ACTIVE;
                return option.activity;
            }
        }
        
        return null;
    }
    
    public void setCommitment(Commitment _commitment) {
        commitment = _commitment;
    }
    
    public Commitment getCommitment() {
        return commitment;
    }
    
    public boolean isToBeMantained() {
        return isToBeMantained;
    }
    
    public void setIsToBeMantained(boolean isToBeMantained) {
        this.isToBeMantained = isToBeMantained;
    }
    
    public void reset() {
        options.clear();
        commitment = null;
    }
}

