package com.agentfactory.agentspeak.interpreter;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Set;

import com.agentfactory.agentspeak.interpreter.mentalstate.ASEvent;
import com.agentfactory.agentspeak.interpreter.mentalstate.AchievementFormula;
import com.agentfactory.agentspeak.interpreter.mentalstate.Intention;
import com.agentfactory.agentspeak.interpreter.mentalstate.PartialPlan;
import com.agentfactory.agentspeak.interpreter.mentalstate.PlanRule;
import com.agentfactory.clf.interpreter.AbstractLanguageAgent;
import com.agentfactory.clf.interpreter.Action;
import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.interpreter.Sensor;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.plans.IPlanStep;
import com.agentfactory.clf.reasoner.Bindings;
import com.agentfactory.clf.reasoner.IQueryEngine;
import com.agentfactory.clf.reasoner.IQueryable;
import com.agentfactory.clf.reasoner.QueryResult;
import com.agentfactory.clf.reasoner.ResolutionBasedQueryEngine;
import com.agentfactory.clf.util.LogicException;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.util.FIPAHelper;
import com.agentfactory.service.migration.MigrationService;

public class AbstractAgentSpeakAgent extends AbstractLanguageAgent implements IQueryable {
	/*********************************************************************
	 * Core Interpreter Data Structures
	 *********************************************************************/
	protected List<ASEvent> eventQueue;
	protected List<IMPLIES> inferences;
	protected BeliefManager beliefManager;
	protected IQueryEngine queryEngine;
	protected List<Intention> intentions;
	protected Map<String, List<PlanRule>> ruleEventMap;
	protected Set<String> eventFilter;
	protected Map<String, PartialPlan> partialPlanMap;
	
	/*********************************************************************
	 * Current State of Algorithm (i.e. what was selected in the loop
	 *********************************************************************/
	private ASEvent event;
	private List<PlanRule> options;
	private PlanRule option;
	private int indentionIndex;
	
	public AbstractAgentSpeakAgent(String name) {
		super(name);

		// Create an event queue and add an "initialized" event
		// that can be used to hook initial behaviours that should
		// be performed by the agent.
		eventQueue = new LinkedList<ASEvent>();
		
		partialPlanMap = new HashMap<String, PartialPlan>();
		beliefManager = new BeliefManager();
		inferences = new LinkedList<IMPLIES>();
		queryEngine = new ResolutionBasedQueryEngine();
		queryEngine.addSource(this);
		queryEngine.addSource(beliefManager);
		eventFilter = new HashSet<String>();
		indentionIndex = -1;
		
		ruleEventMap = new HashMap<String, List<PlanRule>>();
		
		intentions = new LinkedList<Intention>();
		
		installBasicAPI();
		
		// Remove the default inbox sensor because it is incompatible
		// with the AS event model - all messages MUST become events
		this.sensorMap.remove(".inbox:0");
	}
	
	@Override
	public void execute() {
		this.retrieveNewMessages();
//		System.out.println("\n\n============================================");
		beliefManager.copyAndWipePercepts();
		this.senseEnvironment();
		beliefManager.manage();
		for (ASEvent event : beliefManager.events()) {
			if (eventFilter.contains(event.toSignature())) eventQueue.add(event);
		}
//		System.out.println("============================================\n");
		
		for (IMessage message: this.getInbox()) {
        	this.addEvent(new ASEvent(ASEvent.ADDITION, Utilities.factory.createFormula("message(" + message.getPerformative().toLowerCase() + "," +
        			FIPAHelper.toFOSString(message.getSender()) + "," +
        			message.getContent() + ")")));
        }

		this.getOutbox().clear();
		
		// Part B: Handle New Events (if one exists)
		event = getNextEvent();
		
		// NOTE: Here, we check the migration state. If we are doing a gracefull migration
		//       then we look only for goal events (as these are refinements of existing
		//       intentions)
		if (migrationState.migrate && migrationState.type == MigrationState.GRACEFULL) {
			while ((event != null) && (!AchievementFormula.class.isInstance(event.content()))) {
				event = getNextEvent();
			}
		}
		
		if (event != null) {
			// 1. Generate a set of options
			options = generateOptions(event);
				
			// 2. Select an option & commit to it
			option = selectOption(options);
			if (option != null) {
				if (event.getSource() != null) {
					Intention commitment = (Intention) event.getSource();
					commitment.addInitialPlan(option.getPlan());
					commitment.resume();
				} else {	
					intentions.add(new Intention(event, option));
				}
			}
		}
		
		// Part C: Handle Existing Commitments (if any exist)
		if (!intentions.isEmpty()) {
			int N = intentions.size();
//			int N = 1;
			// Process N intentions per iteration...
			for (int i=0; i<N; i++) {
				Intention intention = selectIntention();
				if (intention != null) {
					QueryResult res = this.queryAll(intention.getCondition());
					if (res.found) {
						IPlanStep planStep = intention.getNextActivity();
						if (planStep == null) {
							// No Action => Commitment is completed
							intentions.remove(intention);
						} else {
							planStep.handle(this, intention);
						}
					} else {
						intention.fail();
					}
					
					if (intention.isFailed()) {
						if (!intention.rollback()) {
							intentions.remove(intention);
						}
					}
				}
			}
		} 

		if (migrationState.migrate) {
			// If the migration flag is set, then check which type of
			// migration is to occur and act if the conditions are right
			// NOTE: here we assume that the migration service is bound
			//       to the agent
			if ((migrationState.type == MigrationState.IMMEDIATE) ||
				(migrationState.type == MigrationState.GRACEFULL && eventQueue.isEmpty() && intentions.isEmpty())) {
				
				MigrationService ms = (MigrationService) getPlatformService(MigrationService.NAME);
				if (ms == null) {
					addBelief((Predicate) Utilities.factory.createFormula("migrationFailure(noMigrationService)"));
				} else {
					ms.sendAgent(this, migrationState.host, migrationState.port);
				}
				
				migrationState.migrate = false;
			}
		}

		// Signal the end of the iteration...
        this.endOfIteration();
        event = null;
        option = null;
	}

	public PartialPlan getPartialPlan(Predicate predicate) {
		return partialPlanMap.get(CoreUtilities.createSignature(predicate));
	}
	
	public Map<String, PartialPlan> getPartialPlans() {
		return partialPlanMap;
	}
	
	public QueryResult queryAll(IFormula formula) {
		return queryEngine.queryAll(formula);
	}

	@Override
	/**
	 * Implement response to lack of an action
	 */
	public void noSuchAction(Predicate activity) {
		String a = CoreUtilities.presenter.toString(activity);
		System.out.println("Unknown Action Call: " + a);
		Predicate predicate = (Predicate) CoreUtilities.factory.createFormula("unknownActivity(" + a + ")[source(percept)]");
		this.addEvent(new ASEvent(ASEvent.ADDITION, predicate));
	}
	
	public ASEvent getEvent() {
		return event;
	}
	
	public List<PlanRule> getOptions() {
		return options;
	}
	
	public PlanRule getSelected() {
		return option;
	}
	
	public List<Predicate> getBeliefs() {
		return beliefManager.beliefs();
	}
	
	public List<ASEvent> getEventQueue() {
		return eventQueue;
	}
	
	public List<Intention> getCommitments() {
		return intentions;
	}
	
	public List<PlanRule> getRules() {
		List<PlanRule> list = new LinkedList<PlanRule>();
		for(List<PlanRule> rules: ruleEventMap.values()) {
			list.addAll(rules);
		}
		return list;
	}
	
	public Collection<Action> getActions() {
		return actionMap.values();
	}

	public Collection<Sensor> getSensors() {
		return sensorMap.values();
	}

	@Override
	public String getType() {
		return "AgentSpeak";
	}

	@Override
	public void initialise(String data) {
		if (data.startsWith("+")) {
			Predicate event = (Predicate) Utilities.factory.createFormula(data.substring(1));
			addEvent(new ASEvent(ASEvent.ADDITION, event));
		} else if (data.startsWith("-")) {
			Predicate event = (Predicate) Utilities.factory.createFormula(data.substring(1));
			addEvent(new ASEvent(ASEvent.REMOVAL, event));
		} else {
			try {
				addBelief((Predicate) Utilities.factory.createFormula(data+"[source(self)]"));
			} catch (LogicException le) {
				System.out.println("[" + getName() + "] Initialisation Error: " + le.getMessage());
			}
		}
	}

	@Override
	public void update(Observable arg0, Object arg1) {
		// TODO Auto-generated method stub
	}
	
	public void addBelief(Predicate belief) {
		beliefManager.adoptBelief(belief, null);
	}

	public void addBelief(Predicate belief, Object source) {
		beliefManager.adoptBelief(belief, source);
	}
	
	public void removeBelief(Predicate belief) {
		beliefManager.dropBelief(belief, null);
	}
	
	public void removeBelief(Predicate belief, Object source) {
		beliefManager.dropBelief(belief, source);
	}
	
	public void addEvent(ASEvent event) {
		String sig = event.toSignature();
		if (eventFilter.contains(sig)) eventQueue.add(event);
	}
	
	protected void addPlanRule(PlanRule rule) {
		List<PlanRule> list = ruleEventMap.get(rule.getTriggeringEvent().toSignature());
		if (list == null) {
			list = new LinkedList<PlanRule>();
			ruleEventMap.put(rule.getTriggeringEvent().toSignature(), list);
		}
		list.add(rule);
		eventFilter.add(rule.getTriggeringEvent().toSignature());
	}
	
	protected void addInference(IMPLIES inference) {
		inferences.add(inference);
	}
	
	public List<IMPLIES> getInferences() {
		return inferences;
	}
	
	public IQueryEngine getQueryEngine(){
		return queryEngine;
	}

	@Override
	public boolean hasType(String type) {
		return true;
	}

	@Override
	public List<IFormula> query(IFormula query) {
		List<IFormula> list = new LinkedList<IFormula>();
		if (Predicate.class.isInstance(query)) {
			Predicate predicate = (Predicate) query;
			for (IMPLIES inference:inferences) {
				if (inference.result().matches(predicate)) {
					list.add(inference);
				}
			}
		}
		return list;
	}

	private ASEvent getNextEvent() {
		if (eventQueue.isEmpty()) return null;
		
		return eventQueue.remove(0);
	}
	
	private List<PlanRule> generateOptions(ASEvent event) {
		List<PlanRule> list = ruleEventMap.get(event.toSignature());

		List<PlanRule> options = new LinkedList<PlanRule>();
		if (list != null) {
			for (PlanRule rule: list) {
				if (rule.getTriggeringEvent().matches(event)) {
					Bindings set = rule.getTriggeringEvent().unifyWith(event);
					if (set != null) {
						options.add(rule.apply(set));
					}
				}
			}
		}
		return options;
	}

	private PlanRule selectOption(List<PlanRule> options) {
		for (PlanRule option: options) {
			// Check that the context of the plan rule is satisfied and
			// create a commitment for only the set of bindings that
			// were identified.
			QueryResult result = queryEngine.query(option.getContext());
			if (result.found) {
				if (result.bindings.isEmpty()) {
					return option;
				} else {
					return option.apply(result.bindings.get(0));
				}
			}
		}
		return null;
	}
	
	private Intention selectIntention() {
		// Assume commitments is not empty
		int index = indentionIndex;
		do {
			indentionIndex = (indentionIndex+1) % intentions.size();
		} while ((indentionIndex != index) && intentions.get(indentionIndex).isSuspended());
		
		if (intentions.get(indentionIndex).isSuspended()) return null;
		
		return intentions.get(indentionIndex);
	}
}
