package com.agentfactory.agentspeak.interpreter;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.agentspeak.interpreter.mentalstate.ASEvent;
import com.agentfactory.agentspeak.interpreter.mentalstate.AgentSpeakBasicStep;
import com.agentfactory.agentspeak.interpreter.mentalstate.PlanRule;
import com.agentfactory.clf.interpreter.Action;
import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.interpreter.Module;
import com.agentfactory.clf.interpreter.Sensor;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.Variable;
import com.agentfactory.clf.plans.AssignmentStep;
import com.agentfactory.clf.plans.BlockStep;
import com.agentfactory.clf.plans.DurativeStep;
import com.agentfactory.clf.plans.ForEachStep;
import com.agentfactory.clf.plans.IPlanStep;
import com.agentfactory.clf.plans.IfStep;
import com.agentfactory.clf.plans.TryRecoverStep;
import com.agentfactory.clf.plans.WaitStep;
import com.agentfactory.clf.plans.WhileStep;


public class MigrationBasedAgentSpeakAgent extends AbstractAgentSpeakAgent {
	private class PlanState {
		public PlanState(String[] bits, int index) {
			this.bits = bits;
			this.index = index;
		}
		
		public String[] bits;
		public int index;
	}
	
	public MigrationBasedAgentSpeakAgent(String name) {
		super(name);
		eventQueue.add(new ASEvent(ASEvent.ADDITION, (Predicate) Utilities.factory.createFormula("migrated")));
	}

	public void construct(AgentSpeakMigrationSnapshot snapshot) {
		for(String belief: snapshot.beliefs) {
			addBelief(belief);
		}
		
		for (String inference:snapshot.inferences) {
			addInference((IMPLIES) CoreUtilities.factory.createFormula(inference));
		}
		for (String event:snapshot.events) {
			eventQueue.add(createASEvent(event));
		}
		
		for (String id: snapshot.modules.keySet()) {
			String cl = snapshot.modules.get(id);
			try {
				addModule(id, (Module) Class.forName(cl).newInstance());
			} catch (Exception e) {
				addBelief("missingModule(" + id + "," + cl+ ")");
			}
		}

		for (String _id: snapshot.actions.keySet()) {
			Predicate id = (Predicate) CoreUtilities.factory.createFormula(_id);
			String cl = snapshot.actions.get(id);
			try {
				Action action = (Action) Class.forName(cl).newInstance();
				action.setId(id);
				action.setAgent(this);
				actionMap.put(Utilities.createSignature(id), action);
			} catch (Exception e) {
				addBelief("missingAction(" + id + "," + cl+ ")");
			}
		}
		
		for (String _id: snapshot.sensors.keySet()) {
			Predicate id = (Predicate) CoreUtilities.factory.createFormula(_id);
			String cl = snapshot.sensors.get(id);
			try {
				Sensor sensor = (Sensor) Class.forName(cl).newInstance();				
				sensor.setId(id);
				sensor.setAgent(this);
				sensorMap.put(Utilities.createSignature(id), sensor);
			} catch (Exception e) {
				addBelief("missingSensor(" + id + "," + cl+ ")");
			}
		}
		
		for (AgentSpeakMigrationSnapshot.Rule rule : snapshot.rules ) {
			List<IPlanStep> plan = new LinkedList<IPlanStep>();
			
			for (String stp : rule.plan) {
				plan.add(createPlanStep(new PlanState(stp.split(" "), 0)));
			}
			
			addPlanRule(
					new PlanRule((Predicate) Utilities.factory.createFormula(rule.label),
								createASEvent(rule.event),
								Utilities.factory.createFormula(rule.context),
								Utilities.factory.createFormula("true"), // TODO: Fix this!
								plan));
		}
	}

	private IPlanStep createPlanStep(PlanState state) {
		String cmd = state.bits[state.index++];
		if (cmd.equals("BS")) {
			return new AgentSpeakBasicStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])));
		} else if (cmd.equals("AS")) {
			return new AssignmentStep((Variable) Utilities.factory.createTerm(state.bits[state.index++]), Utilities.factory.createTerm(revertSpaces(state.bits[state.index++])), state.bits[state.index++].equals("yes"));
		} else if (cmd.equals("WH")) {
			return new WhileStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])), createPlanStep(state));
		} else if (cmd.equals("FE")) {
			return new ForEachStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])), createPlanStep(state));
		} else if (cmd.equals("DU")) {
			return new DurativeStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])));
		} else if (cmd.equals("IF")) {
			return new IfStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])), createPlanStep(state), createPlanStep(state));
		} else if (cmd.equals("TR")) {
			return new TryRecoverStep(createPlanStep(state), createPlanStep(state));
		} else if (cmd.equals("WT")) {
			return new WaitStep(Utilities.factory.createFormula(revertSpaces(state.bits[state.index++])));
		} else if (cmd.equals("SB")) {
			List<IPlanStep> list = new LinkedList<IPlanStep>();
			while (!state.bits[state.index].equals("EB")) {
				list.add(createPlanStep(state));
			}
			state.index++;
			return new BlockStep(list);
		}
		
		System.out.println("Unknown plan part: " + cmd);
		return null;
	}
	
	private ASEvent createASEvent(String evt) {
		if (evt.startsWith("+")) {
			return new ASEvent(ASEvent.ADDITION, Utilities.factory.createFormula(evt.substring(1)));
		} else if (evt.startsWith("-")) {
			return new ASEvent(ASEvent.REMOVAL, Utilities.factory.createFormula(evt.substring(1)));
		}
		
		return null;
	}

	private void addBelief(String belief) {
		beliefManager.adoptBelief((Predicate) Utilities.factory.createFormula(belief), null);
	}
	
	private String revertSpaces(String str) {
		return str.replace("%20", " ");
	}
}
