package com.agentfactory.clf.plans;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.lang.ITerm;
import com.agentfactory.clf.lang.Variable;
import com.agentfactory.clf.reasoner.Bindings;

public class PlanStack {
	private List<IPlanStep> stack;
	private IPlanStep current;
	private int state;
	private List<Bindings> bindings = new LinkedList<Bindings>();
	private List<Integer> popPoints = new LinkedList<Integer>();
	private Map<Integer, List<IPlanStep>> recoveryPlans = new HashMap<Integer, List<IPlanStep>>();
	
	public static final int STATE_ACTIVE						 = 0;
	public static final int STATE_SUSPENDED						 = 1;
	public static final int STATE_FAILED						 = 2;
	public static final int STATE_SUCCEEDED						 = 3;

	
	public PlanStack() {
		stack = new LinkedList<IPlanStep>();
		bindings.add(new Bindings());
		popPoints.add(0);
	}

	public String toString() {
		String output = "";
		for (Bindings b: bindings) {
			output += CoreUtilities.presenter.toString(b) + " / ";
		}
		
		return output;
	}
	
	public String stateString() {
		switch (state) {
		case STATE_ACTIVE:
			return "ACTIVE";
		case STATE_SUSPENDED:
			return "SUSPENDED";
		case STATE_FAILED:
			return "FAILED";
		case STATE_SUCCEEDED:
			return "SUCCEEDED";
		}
		return "NO STATE";
	}
	
	public List<IPlanStep> getStack() {
		return stack;
	}
	
	public IPlanStep getCurrent() {
		return current;
	}
	
	public void addInitialPlan(List<IPlanStep> list) {
		for (int i=list.size()-1; i>=0; i--) {
			stack.add(0, list.get(i));
		}
	}
	
	public void addPlan(List<IPlanStep> subplan) {
		pushBindings();
		popPoints.add(stack.size());
		for (int i=subplan.size()-1; i>=0; i--) {
			stack.add(0, subplan.get(i));
		}
	}
	
	public void back() {
		stack.add(0, current);
	}
	
	public IPlanStep getNextActivity() {
		if (stack.isEmpty()) {
			// We have completed the plan...
			state = PlanStack.STATE_SUCCEEDED;			
			current = null;
		} else {
			if (popPoints.get(popPoints.size()-1) == stack.size()) {
				popPoints.remove(popPoints.size()-1);
				popBindings();
			}
			current = stack.remove(0);
		}
		
		return current;
	}
	
	public void suspend() {
		state = PlanStack.STATE_SUSPENDED;
	}
	
	public void resume() {
		state = PlanStack.STATE_ACTIVE;
	}
	
	public void fail() {
		state = PlanStack.STATE_FAILED;
	}
	
	public boolean isSuspended() {
		return state == PlanStack.STATE_SUSPENDED;
	}
	
	public boolean isActive() {
		return state == PlanStack.STATE_ACTIVE;
	}

	public boolean isFailed() {
		return state == PlanStack.STATE_FAILED;
	}

	public boolean isSucceeded() {
		return state == PlanStack.STATE_SUCCEEDED;
	}
	
	public void pushBindings() {
		bindings.add(new Bindings());
	}
	
	public Bindings popBindings() {
		return bindings.remove(bindings.size()-1);
	}
	
	public void updateBindings(Bindings bindings) {
		Iterator<ITerm[]> it = bindings.iterator();
		while (it.hasNext()) {
			ITerm[] terms = it.next();
			updateBinding((Variable) terms[0], terms[1]);
		}
	}
	
	
	public void updateBinding(Variable variable, ITerm term) {
		int index = 0;
		while (index < bindings.size()) {
			Bindings b = bindings.get(index++);
			ITerm current = b.getBinding(variable);
			if (current != null) {
				b.addBinding(variable, term);
				return;
			}
		}
		
		bindings.get(bindings.size()-1).addBinding(variable, term);
	}
	
	public Bindings bindings() {
		Bindings current = new Bindings();
		for (Bindings b:bindings) {
			current = current.merge(b);
		}
		return current;
	}

	/**
	 * Adds a recovery plan to the commitment. This plan is linked to the stack size when it
	 * is added. The plan is triggered in the case where the commitment has failed and the
	 * stack is being rolled back. When the stack level corresponds to a recovery plan, recovery
	 * occurs.
	 * 
	 * @param plan
	 */
	public void addRecoveryPlan(List<IPlanStep> plan) {
		recoveryPlans.put(stack.size(), plan);
	}
	
	public boolean rollback() {
		while (!stack.isEmpty() && (recoveryPlans.get(stack.size()) == null)) {
			getNextActivity();
		}

		List<IPlanStep> plan = recoveryPlans.remove(stack.size());
		if (plan == null) return false;

		popBindings();
		addPlan(plan);
		
		this.resume();
		
		return true;
	}
}
