package com.agentfactory.clf.reasoner;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.clf.lang.AND;
import com.agentfactory.clf.lang.Equality;
import com.agentfactory.clf.lang.GreaterThan;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.LessThan;
import com.agentfactory.clf.lang.NOT;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.Unifier;
import com.agentfactory.clf.util.DefaultLogicFactory;
import com.agentfactory.clf.util.DefaultLogicPresenter;

public class ForwardChainer implements IQueryEngine {
	List<IQueryable> sources;
	DerivedSource derived;
	List<IMPLIES> rules;
	
	private class Entry {
		Bindings bindings;
		List<IFormula> formulae;
		
		public Entry(Entry entry, Bindings bindings) {
			this();
			for (IFormula formula: entry.formulae) {
				this.formulae.add(formula.apply(bindings));
			}
			
			this.bindings = entry.bindings.merge(bindings);
		}
		
		public Entry(IFormula formula) {
			this();
			add(formula);
		}
		
		public Entry() {
			bindings = new Bindings();
			formulae = new LinkedList<IFormula>();
		}
		
		public void add(IFormula formula) {
			if (Predicate.class.isInstance(formula) || NOT.class.isInstance(formula)) {
				formulae.add(formula);
			} else if (AND.class.isInstance(formula)) {
				for (IFormula f: formula.asList()) {
					add(f);
				}
			} else {
				System.out.println("WARNING: Tried to add invalid formula");
			}
		}
		
		public String toString() {
			String output = "formula stack: ";
			for (IFormula formula: formulae) {
				output += presenter.toString(formula) + ", ";
			}
			return output;
		}
	}
	
	public ForwardChainer() {
		sources = new LinkedList<IQueryable>();
		derived = new DerivedSource();
		sources.add(derived);
		rules = new LinkedList<IMPLIES>();
	}
	
	public void addSource(IQueryable source) {
		sources.add(source);
	}

	public void addRule(IMPLIES rule) {
		rules.add(rule);
	}
	
	public void refresh() {
		derived.clear();
		int size = -1;
		while (size != derived.size()) {
			size = derived.size();
		
			for (IMPLIES rule: rules) {
				List<Bindings> bindings = solve(rule.condition(), true);
//				if(bindings.size() >0){
//				System.out.print("Rule: "+ presenter.toString(rule)+"  bindings found = "+bindings.size()+" of " + derived.size());
//				for(Bindings b:bindings)	if(rule.result().apply(b).getVariables().size() ==0) System.out.print(" "+presenter.toString(b)+", "+ presenter.toString(rule.result().apply(b)));
//				System.out.println();}
				if (bindings != null) {
					for (Bindings binding: bindings) {
						if(rule.result().apply(binding).getVariables().size() ==0)
						derived.add((Predicate) rule.result().apply(binding));
					}
				}
			}
		}
	}
	
	static DefaultLogicFactory factory = new DefaultLogicFactory();
	static DefaultLogicPresenter presenter = new DefaultLogicPresenter();
	
	private List<Bindings> solve(IFormula formula, boolean findAll) {
		List<Bindings> bindings = new LinkedList<Bindings>();
		List<Entry> solutions = new LinkedList<Entry>();
		List<Entry> options = new LinkedList<Entry>();
		options.add(new Entry(formula));
//		System.out.println("[solve] start");
		while (!options.isEmpty()) {
//			System.out.println("Loopity-loop");
			Entry option = options.remove(0);
			if (option.formulae.isEmpty()) {
				if (!findAll) {
					bindings.add(option.bindings);
					return bindings;
				}
				solutions.add(option);
			} else {
				IFormula goal = option.formulae.remove(0);
				
				if (Predicate.class.isInstance(goal)) {
//					System.out.println("[solve] processing: " + CoreUtilities.presenter.toString(goal));
					for (IQueryable source: sources) {
						if (source.hasType(goal.type())) {
							for (IFormula form: source.query(goal)) {
//								System.out.println("[solve] testing: " + CoreUtilities.presenter.toString(form));
								Bindings b = Unifier.unifyFormulae((Predicate) goal, (Predicate) form);
								if (b != null) {
//									System.out.println("[solve] binding: " + CoreUtilities.presenter.toString(b));
									Entry e = new Entry(option, b);
									options.add(e);
								}
							}
						}
					}
				} else if (NOT.class.isInstance(goal)) {
					List<Bindings> b = solve(((NOT) goal).formula(), false);
					if (b == null) {
						options.add(option);
					}
				} else if (Equality.class.isInstance(formula) || LessThan.class.isInstance(formula) || GreaterThan.class.isInstance(formula)) {
	        		if (formula.isTrue()) options.add(option);
	        	}

			}
		}
		
		if (solutions.isEmpty()) return null;
		
//		System.out.println("[solve] binding generation");
		for (Entry entry: solutions) {
			if (!bindings.contains(entry.bindings)) bindings.add(entry.bindings);
		}
		
//		System.out.println("[solve] end");
		return bindings;
	}
	
	public static class DerivedSource implements IQueryable {
		List<Predicate> formulae = new LinkedList<Predicate>();
		
		@Override
		public boolean hasType(String type) {
			return true;
		}
		
		public List<Predicate> formulae() {
			return formulae;
		}
		
		public void add(Predicate predicate) {
			if (!formulae.contains(predicate)) formulae.add(predicate);
		}
		
		public int size() {
			return formulae.size();
		}

		@Override
		public List<IFormula> query(IFormula query) {
			List<IFormula> list = new LinkedList<IFormula>();
			
			Predicate p = (Predicate) query;
			for (Predicate pred: formulae) {
				if (p.matches(pred)) {
					list.add(pred);
				}
			}
			return list;
		}
		public void print(){
			List<String> list = new LinkedList<String>();
			for(Predicate p : formulae){
				list.add(presenter.toString(p));
			}
			System.out.println(list);
		}
		
		public void clear() {
			formulae.clear();
		}

		public List<Predicate> getDerivedBeliefs() {
			List<Predicate> l = new LinkedList<Predicate>();
			for(Predicate p : formulae)
				l.add(p);
			return l;
		}
	}
	
	public static class TestSource implements IQueryable {
		List<Predicate> formulae = new LinkedList<Predicate>();
		
		public TestSource() {
			formulae.add((Predicate)factory.createFormula("is(A,B)"));
			formulae.add((Predicate)factory.createFormula("is(B,C)"));
			formulae.add((Predicate)factory.createFormula("is(C,D)"));
			formulae.add((Predicate)factory.createFormula("on(A,B)"));
			formulae.add((Predicate)factory.createFormula("on(B,C)"));
			formulae.add((Predicate)factory.createFormula("on(E,B)"));
			formulae.add((Predicate)factory.createFormula("on(C,D)"));
			formulae.add((Predicate)factory.createFormula("clear(B)"));
			formulae.add((Predicate)factory.createFormula("clear(D)"));
		}
		
		@Override
		public boolean hasType(String type) {
			return type.equals("raw-logic");
		}

		@Override
		public List<IFormula> query(IFormula query) {
			List<IFormula> list = new LinkedList<IFormula>();
			
			Predicate p = (Predicate) query;
			for (Predicate pred: formulae) {
				if (p.matches(pred)) {
					list.add(pred);
				}
			}
			// TODO Auto-generated method stub
			return list;
		}
	}
	
	public static void main(String[] args) {
		ForwardChainer chainer = new ForwardChainer();
		chainer.addSource(new TestSource());
		chainer.addRule((IMPLIES)factory.createFormula("on(?x, ?y) & ~clear(?y) => available(?x)"));
//		chainer.addRule((IMPLIES)factory.createFormula("is(?x, ?y) & is(?y,?z) => is(?x,?z)"));
//		chainer.addRule((IMPLIES)factory.createFormula("is(?x, ?y) => on(?x,?y)"));
		chainer.refresh();
		QueryResult result = chainer.queryAll(factory.createFormula("available(?x)"));
		if (result.found) {
			for (Bindings b:result.bindings) {
				System.out.println(presenter.toString(b));
			}
		} else {
			System.out.println("No such solution");
		}
		
	}

	@Override
	public QueryResult query(IFormula formula) {
		QueryResult result = new QueryResult();
		result.bindings = solve(formula, false);
		result.found = result.bindings != null;
		return result;
	}

	@Override
	public QueryResult queryAll(IFormula formula) {
		QueryResult result = new QueryResult();
		result.bindings = solve(formula, true);
		result.found = result.bindings != null;
		return result;
	}

	public List<Predicate> getDerivedBeliefs() {
		return derived.getDerivedBeliefs();
	}

	public List<IMPLIES> getRules() {
		return rules;
	}
}
