package com.agentfactory.clf.util;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.clf.lang.AND;
import com.agentfactory.clf.lang.Addition;
import com.agentfactory.clf.lang.Bracket;
import com.agentfactory.clf.lang.Contains;
import com.agentfactory.clf.lang.Division;
import com.agentfactory.clf.lang.Equality;
import com.agentfactory.clf.lang.Function;
import com.agentfactory.clf.lang.GreaterThan;
import com.agentfactory.clf.lang.Head;
import com.agentfactory.clf.lang.IFormula;
import com.agentfactory.clf.lang.IMPLIES;
import com.agentfactory.clf.lang.ITerm;
import com.agentfactory.clf.lang.Inequality;
import com.agentfactory.clf.lang.IntegerTerm;
import com.agentfactory.clf.lang.LessThan;
import com.agentfactory.clf.lang.ListTerm;
import com.agentfactory.clf.lang.Merge;
import com.agentfactory.clf.lang.Modulo;
import com.agentfactory.clf.lang.Multiplication;
import com.agentfactory.clf.lang.NOT;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.RealTerm;
import com.agentfactory.clf.lang.StringTerm;
import com.agentfactory.clf.lang.Subtraction;
import com.agentfactory.clf.lang.Tail;
import com.agentfactory.clf.lang.Variable;

/**
 * This class implements support for converting string representations
 * of logical formulae based on the default logic syntax used in
 * Agent Factory into objects:
 * - and (&)
 * - variable (?)
 * - not (!)
 * - implication (=>)
 * 
 * To use a different syntax all you need to do is create a new Logic
 * Presenter and a new Logic Factory implementation.
 * 
 * To extend the basic syntax to include modal/temporal operators, ...
 * simply subclass this class and override the createCustomFormulae(...)
 * method.
 * 
 * @author rem
 *
 */
public class DefaultLogicFactory implements ILogicFactory {
    public static final ILogicPresenter presenter = new DefaultLogicPresenter();

    @Override
	public IFormula createFormula(String fos) {
		// Precedence:
		// 1. Look for IMPLIES
		// 2. Look for AND
		// 3. Look for NOT
		// 4. User defined operators...
		// 5. Look for formula
		 
		// 1. Assume that there is max one implies sign...
		int index = findIndexOf(fos,"=>");
		if (index > -1) {
			// found an implies sign...
			IFormula left = createFormula(fos.substring(0, index).trim());
			IFormula right = createFormula(fos.substring(index+2).trim());
			return new IMPLIES(left, right); // left => right
		}
		
		// 2. Look for first AND
		index = findIndexOf(fos,"&");
		if (index > -1) {
			// Split the sentence into a literal + the rest and recurse
			List<IFormula> list = new LinkedList<IFormula>();
			String temp = fos;
			while (index > -1) {
				list.add(createFormula(temp.substring(0, index).trim()));
				temp = temp.substring(index+1).trim();
				int index2 = findIndexOf(temp,"&");
				if (index2 == -1) {
					list.add(createFormula(temp));
				}
				index = index2;
			}
			return new AND(list); // left => right
		}
		
		if (fos.startsWith("~")) {
			return new NOT(createFormula(fos.substring(1)));
		}
				
		// 2. Look for first AND
		index = findIndexOf(fos,"==");
		if (index > -1) {
			// Split the sentence into a literal + the rest and recurse
			ITerm left = createTerm(fos.substring(0, index).trim());
			ITerm right = createTerm(fos.substring(index+2).trim());
			return new Equality(left, right); // left => right
		}
		
		index = findIndexOf(fos,"!=");
		if (index > -1) {
			// Split the sentence into a literal + the rest and recurse
			ITerm left = createTerm(fos.substring(0, index).trim());
			ITerm right = createTerm(fos.substring(index+2).trim());
			return new Inequality(left, right); // left => right
		}
		
		index = findIndexOf(fos,">");
		if (index > -1) {
			// Split the sentence into a literal + the rest and recurse
			ITerm left = createTerm(fos.substring(0, index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			return new GreaterThan(left, right); // left => right
		}
		
		index = findIndexOf(fos,"<");
		if (index > -1) {
			// Split the sentence into a literal + the rest and recurse
			ITerm left = createTerm(fos.substring(0, index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			return new LessThan(left, right); // left => right
		}

		if (fos.startsWith("$(")) {
			// Contains formula...
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				throw new LogicException("Unmatched brackets when parsing formula: " + fos);
			}
			String arguments = fos.substring(2, index2);
			index = findIndexOf(arguments, ",");
			ITerm term1 = createTerm(arguments.substring(0,index).trim());
			ITerm term2 = createTerm(arguments.substring(index+1).trim());
			return new Contains(term1, term2);
		}
		IFormula formula = createCustomFormula(fos);
		if (formula != null) return formula;
		
		if (fos.endsWith("]")) {
			// Possible label
			index = fos.lastIndexOf("[");
			String labelString = fos.substring(index+1, fos.length()-1).trim();
			List<Predicate> labels = new LinkedList<Predicate>();
			
			Predicate pred = (Predicate) createFormula(fos.substring(0, index));
			List<ITerm> terms = new LinkedList<ITerm>();
			for (ITerm term:pred) {
				terms.add(term);
			}
			index = findIndexOf(labelString, ",");
			while (index > -1) {
				labels.add((Predicate) createFormula(labelString.substring(0,index).trim()));
				labelString = labelString.substring(index+1).trim();
				index = findIndexOf(labelString, ",");
			}
			labels.add((Predicate) createFormula(labelString.trim()));
			
			return new Predicate(pred.functor(), terms, labels);
		}
		
		// Okay, so we have a basic predicate, so lets deal with it...
		index = fos.indexOf('(');
		if (index == -1) {
			return new Predicate(fos);
		} else if (index != 0){
			String functor = fos.substring(0,index);
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				throw new LogicException("Unmatched brackets when parsing formula: " + fos);
			}
			String arguments = fos.substring(index+1, index2);
			
			List<ITerm> terms = new LinkedList<ITerm>();
			
			index = findIndexOf(arguments, ",");
			while (index > -1) {
				terms.add(createTerm(arguments.substring(0,index).trim()));
				arguments = arguments.substring(index+1).trim();
				index = findIndexOf(arguments, ",");
			}

			terms.add(createTerm(arguments));
			return new Predicate(functor, terms);
		}else {
			String sub = fos.substring(index+1, fos.lastIndexOf(')'));
			return new Bracket(createFormula(sub));
		}
	}
	
	/**
	 * This is method a hook to allow the introduction of custom formulae such
	 * as modal operators (BELIEF, GOAL, ...) or temporal operator (ALWAYS,
	 * UNTIL, ...)
	 *  
	 * @param fos
	 * @return
	 */
	protected IFormula createCustomFormula(String fos) {
		return null;
	}

	private static int findIndexOf(String text, String pattern) {
		int index = 0;
		int pIndex = 0;
		char pch = pattern.charAt(pIndex++);
		List<Character> stack = new LinkedList<Character>();
		while (index < text.length()) {
			char tch = text.charAt(index++);
			if (tch == '(') {
				if (stack.isEmpty()) {
					stack.add(tch);
				} else if (stack.get(stack.size()-1) != '\"') {
					stack.add(tch);
				}
			} else if (tch == '[') {
				if (stack.isEmpty()) {
					stack.add(tch);
				} else if (stack.get(stack.size()-1) != '\"') {
					stack.add(tch);
				}
			} else if (tch == ']') {
				if (stack.get(stack.size()-1) == '\"') {
					// Ignore as this is part of a string
				} else if (stack.get(stack.size()-1) == '[') {
					stack.remove(stack.size()-1);
				} else {
					throw new LogicException("Unmatched brackets when parsing text: " + text);
				}
			} else if (tch == ')') {
				if (stack.get(stack.size()-1) == '\"') {
					// Ignore as this is part of a string
				} else if (stack.get(stack.size()-1) == '(') {
					stack.remove(stack.size()-1);
				} else {
					throw new LogicException("Unmatched brackets when parsing text: " + text);
				}
			} else if (tch == '\"') {
				if (stack.isEmpty()) {
					stack.add(tch);
				} else if (stack.get(stack.size()-1) == '\"') {
					stack.remove(stack.size()-1);
				} else {
					stack.add(tch);
				}
			}

			if ((tch == pch) && stack.isEmpty()) {
				if (pIndex == pattern.length()) {
					return index-pIndex;
				}
				pch = pattern.charAt(pIndex++);
			} else {
				pIndex=1;
				pch = pattern.charAt(0);
			}
		}
		
		if (!stack.isEmpty()) {
			throw new LogicException("Unmatched brackets when parsing text: " + text);
		}
		return -1;
	}
	
	@Override
	public ITerm createTerm(String fos) {
		// TODO: Typically, + and - have the same precedence. Same for / * %.
		int index;
		index = findIndexOf(fos, "+");
		if ((index > -1) && (fos.indexOf('(') == -1)) {
			ITerm left = createTerm(fos.substring(0,index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			if (left.hasTerms() || right.hasTerms()) {
				// Warning: Addition should only contain constants and variables!
				return null;
			}
			return new Addition(new ITerm[] {left, right});
		}

		index = findIndexOf(fos, "-");
		if ((index > -1) && (fos.indexOf('(') == -1)) {
			if (isVariableOrNumber(fos, index)) {
				ITerm left = createTerm(fos.substring(0,index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
				if (left.hasTerms() || right.hasTerms()) {
					return null;
				}
				return new Subtraction(new ITerm[] {left, right});
			}
		}
		index = findIndexOf(fos, "%");
		if ((index > -1) && (fos.indexOf('(') == -1)) {
			ITerm left = createTerm(fos.substring(0,index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			if (left.hasTerms() || right.hasTerms()) {
				// Warning: Addition should only contain constants and variables!
				return null;
			}
			return new Modulo(new ITerm[] {left, right});
		}
		index = findIndexOf(fos, "*");
		if ((index > -1) && (fos.indexOf('(') == -1)) {
			ITerm left = createTerm(fos.substring(0,index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			if (left.hasTerms() || right.hasTerms()) {
				// Warning: Addition should only contain constants and variables!
				return null;
			}
			return new Multiplication(new ITerm[] {left, right});
		}
		index = findIndexOf(fos, "/");
		if ((index > -1) && (fos.indexOf('(') == -1)) {
			ITerm left = createTerm(fos.substring(0,index).trim());
			ITerm right = createTerm(fos.substring(index+1).trim());
			if (left.hasTerms() || right.hasTerms()) {
				// Warning: Addition should only contain constants and variables!
				return null;
			}
			return new Division(new ITerm[] {left, right});
		}

		if (fos.startsWith("\"") && fos.endsWith("\"")) {
    		return new StringTerm(fos.substring(1, fos.length()-1));
//    		return new Function(fos);
    	}

    	if (fos.startsWith("[") && fos.endsWith("]")) {
    		index=1;
			String arguments = fos.substring(index, fos.length()-1).trim();
			
			List<ITerm> terms = new LinkedList<ITerm>();
			if (!arguments.isEmpty()) {
				
				index = findIndexOf(arguments, ",");
				while (index > -1) {
					terms.add(createTerm(arguments.substring(0,index).trim()));
					arguments = arguments.substring(index+1).trim();
					index = findIndexOf(arguments, ",");
				}
				terms.add(createTerm(arguments));
			}
			return new ListTerm(terms);
    	}
    	
    	
		
		// 1. Check for Variables
		if (fos.startsWith("?")) {
			if (fos.indexOf('(') > -1) {
				// Warning: variables should not have arguments
				return null;
			}
			return new Variable(fos.substring(1));
		}
		
		if ((fos.length() > 4) && fos.substring(0, 4).equalsIgnoreCase("head")) {
			index = fos.indexOf('(');
			if (index == -1) {
				throw new LogicException("Expected ( after head when parsing formula: " + fos);
			}
			
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				throw new LogicException("Unmatched brackets when parsing formula: " + fos);
			}
			return new Head(createTerm(fos.substring(index+1, index2).trim()));
		}
		
		if ((fos.length() > 4) && fos.substring(0, 4).equalsIgnoreCase("tail")) {
			index = fos.indexOf('(');
			if (index == -1) {
				throw new LogicException("Expected ( after tail when parsing formula: " + fos);
			}
			
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				throw new LogicException("Unmatched brackets when parsing formula: " + fos);
			}
			return new Tail(createTerm(fos.substring(index+1, index2).trim()));
		}
		
		if ((fos.length() > 5) && fos.substring(0, 5).equalsIgnoreCase("merge")) {
			index = fos.indexOf('(');
			if (index == -1) {
				throw new LogicException("Expected ( after merge when parsing formula: " + fos);
			}
			
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				throw new LogicException("Unmatched brackets when parsing formula: " + fos);
			}
			
			String args = fos.substring(index+1, index2).trim();
			index = findIndexOf(args, ",");
			ITerm[] term = new ITerm[2];
			term[0] = createTerm(args.substring(0, index).trim());
			term[1] = createTerm(args.substring(index+1).trim());
			return new Merge(term);
		}

		// 2. Check for functions
		index = fos.indexOf("(");
		if (index > -1) {
			String functor = fos.substring(0,index);
			int index2 = fos.lastIndexOf(')');
			if (index2 == -1) {
				// Warning: Unmatched Brackets...
				return null;
			}
			String arguments = fos.substring(index+1, index2);
			
			List<ITerm> terms = new LinkedList<ITerm>();
			
			index = findIndexOf(arguments, ",");
			while (index > -1) {
				terms.add(createTerm(arguments.substring(0,index).trim()));
				arguments = arguments.substring(index+1).trim();
				index = findIndexOf(arguments, ",");
			}
			terms.add(createTerm(arguments));
			return new Function(functor, terms);
		} else {
			try {
				return new IntegerTerm(Integer.parseInt(fos));
			} catch (NumberFormatException nfe) {
				try {
					return new RealTerm(Double.parseDouble(fos));
				} catch (NumberFormatException nfe2) {
				}
			}
		}
		
		// 3. Check for maths operators
		return new Function(fos);
	}
	
	public Predicate convertToPredicate(ITerm term) {
		if (!Function.class.isInstance(term)) return null;
		
		Function function = (Function) term;
		List<ITerm> list = new LinkedList<ITerm>();
		for (int i=0; i<function.size(); i++) {
			list.add(function.termAt(i));
		}
		return new Predicate(function.functor(), list);
	}

	public ITerm convertToTerm(Predicate predicate) {
		List<ITerm> list = new LinkedList<ITerm>();
		for (ITerm term : predicate.terms()) {
			list.add(term);
		}
		return new Function(predicate.functor(), list);
	}

	private boolean isVariableOrNumber(String fos, int index) {
		if (fos.trim().charAt(0) == '?') {
			return true;
		}
		
		try {
			Double.parseDouble(fos.substring(0,index).trim());
		} catch (NumberFormatException re) {
			return false;
		}
		
		return true;
	}
}
