package com.agentfactory.eis;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.agentfactory.clf.interpreter.Action;
import com.agentfactory.clf.interpreter.CoreUtilities;
import com.agentfactory.clf.interpreter.Module;
import com.agentfactory.clf.interpreter.Sensor;
import com.agentfactory.clf.lang.Function;
import com.agentfactory.clf.lang.ITerm;
import com.agentfactory.clf.lang.IntegerTerm;
import com.agentfactory.clf.lang.Predicate;
import com.agentfactory.clf.lang.RealTerm;
import com.agentfactory.clf.lang.StringTerm;

import eis.exceptions.ActException;
import eis.exceptions.NoEnvironmentException;
import eis.exceptions.PerceiveException;
import eis.iilang.Identifier;
import eis.iilang.Numeral;
import eis.iilang.Parameter;
import eis.iilang.Percept;


public class EISControlModule extends Module {
	protected EISService service;
	boolean registered = false;
	boolean associated = false;
	String association = "";
	
	public EISControlModule() {
		addAction("setup(?pid)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String pid = CoreUtilities.presenter.toString(activity.termAt(0));
				agent.bindToPlatformService(pid);
				service = (EISService) agent.getPlatformService(pid);
				addBelief("setup(done)");
				return true;
			}
		});
		addAction("registerAgent", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				service.registerAgent(agent);
				registered = true;
				addBelief("registeredAgent(" + agent.getName() + ")");
				return true;
			}
		});
		addAction("associateEntity(?entity)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				String entity = CoreUtilities.presenter.toString(activity.termAt(0));
				service.associateEntity(agent, entity);
				associated = true;
				association = "associatedWithEntity(" + entity + ")";
				addBelief("associatedWithEntity(" + entity + ")");
				return true;
			}
		});
		addAction("perform(?action)", new Action() {
			@Override
			public boolean execute(Predicate activity) {
				if (Function.class.isInstance(activity.termAt(0))) {
					Function arg = (Function) activity.termAt(0);
					
					String act = arg.functor();
					
					LinkedList<Parameter> list = new LinkedList<Parameter>();
					for (int i=0; i<arg.size(); i++) {
						ITerm term = arg.termAt(i);
						if (RealTerm.class.isInstance(term)) {
							list.add(new Numeral(((RealTerm) term).value()));
						} else if (IntegerTerm.class.isInstance(term)) {
							list.add(new Numeral(((IntegerTerm) term).value()));
						} else if (StringTerm.class.isInstance(term)) {
							list.add(new Identifier(((StringTerm) term).value()));
						} else {
							list.add(new Identifier(CoreUtilities.presenter.toString(term)));
						}
					}
					try {
						Map<String, Percept> percepts = service.performAction(agent, new eis.iilang.Action(act, list));

						for (Percept percept: percepts.values()) {
							addBelief(generateBelief(percept)+"[source(percept)]");
						}
						return true;
					} catch (ActException e) {
						if( e.getType() == ActException.NOTSPECIFIC) {
							System.out.println(agent.getName() + ": Action Failure (NOTSPECIFIC): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.NOTREGISTERED ) {
								System.out.println(agent.getName() + ": Action Failure (NOTREGISTERED): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.NOENTITIES ) {
								System.out.println(agent.getName() + ": Action Failure (NOENTITIES): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.WRONGENTITY ) {
								System.out.println(agent.getName() + ": Action Failure (WRONGENTITY): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.NOTSUPPORTEDBYENTITY ) {
								System.out.println(agent.getName() + ": Action Failure (NOTSUPPORTEDBYENTITY): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.NOTSUPPORTEDBYENVIRONMENT ) {
								System.out.println(agent.getName() + ": Action Failure (NOTSUPPORTEDBYENVIRONMENT): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.NOTSUPPORTEDBYTYPE ) {
								System.out.println(agent.getName() + ": Action Failure (NOTSUPPORTEDBYTYPE): " + CoreUtilities.presenter.toString(activity));
							} else if( e.getType() == ActException.FAILURE ) {
								System.out.println(agent.getName() + ": Action Failure (FAILURE): " + CoreUtilities.presenter.toString(activity));
							}
						addBelief("failedAction(" + CoreUtilities.presenter.toString(activity) + ")");
					} catch (NoEnvironmentException e) {
						System.out.println(agent.getName() + ": Action Failed: " + CoreUtilities.presenter.toString(activity));
						addBelief("failedAction(" + CoreUtilities.presenter.toString(activity) + ")");
					}
				}
				return false;
			}
		});
		addSensor("eisSensor", new Sensor() {
			@Override
			public void perceive() {
				if(registered)
					addBelief("registeredAgent(" + agent.getName() + ")");
				if(associated)
					addBelief(association);
				if (service != null)
					addBelief("setup(done)");
				else 
					return;
				
				try {
					Map<String, Collection<Percept>> percepts = service.getPercepts(agent);
					for (Collection<Percept> list : percepts.values()) {
						for (Percept percept: list) {
							String bel = generateBelief(percept);
							if (bel != null) addBelief(bel);
						}
					}
				} catch (PerceiveException e) {
				} catch (NoEnvironmentException e) {
				}
				List<Percept> percepts = service.getAgentEvents(agent);
				while (percepts!=null && !percepts.isEmpty()) {
					addBelief(generateBelief(percepts.remove(0)));
				}
			}
			
		});
	}
	
	private String generateBelief(Percept percept) {
		if (percept == null) {
//			System.out.println("null percept!");
			return null;
		}
		String fos = percept.getName();
		boolean first = true;
		for (Parameter param: percept.getParameters()) {
			if (first) {
				fos += "(";
				first = false;
			} else {
				fos += ",";
			}
			
			String p = param.toProlog();
//			if (Character.isUpperCase(p.charAt(0))) fos += "?";
			fos += p;
		}
		if (!first) fos += ")";
		return fos;
	}
}
