package com.agentfactory.logic.agent;

import com.agentfactory.logic.lang.FOS;
import com.agentfactory.logic.MentalAgent;
import com.agentfactory.platform.core.IPlatformService;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * All actuator classes must subclass this class.
 * 
 * Actuators are activated as a result of decisions made by the agent
 * interpeter. Each action in a commitment must represent an actuator.
 * 
 * @author Rem Collier
 */
public abstract class Actuator {
   protected MentalAgent agent;
   protected ActivityIdentifier actionIdentifier;
   protected boolean threaded;
   protected Map<String,String> configuration;
   private List<String> ontologies;

   /***************************************************************************
    * CONSTRUCTORS
    ***************************************************************************/
   public Actuator() {
      this.threaded = true;
   }

   /***************************************************************************
    * ACCESSOR METHODS
    ***************************************************************************/
   public void setAgent( MentalAgent agent ) {
      this.agent = agent;
   }

   public boolean isThreaded() {
      return threaded;
   }

   public IPlatformService getService( String name ) {
      return agent.getPlatformService( name );
   }

   public <T extends IPlatformService> List<T> getServicesByClass( Class<T> targetClass ) {
      return agent.getPlatformServicesByClass( targetClass );
   }

   public Module getModuleByClass( String className ) {
      return agent.getModuleByClass( className );
   }

   public Module getModuleByName( String name ) {
      return agent.getModuleByName( name );
   }

   public void destroy() {}

   public void setOntologies( List<String> ontologies ) {
      this.ontologies = ontologies;
   }

   public Belief createBelief( String functor ) {
      FOS term = null;
      Ontology ontology = null;
      Iterator<String> it = ontologies.iterator();
      while ( it.hasNext() ) {
         ontology = agent.getOntologyManager().getOntology( it.next() );
         term = ontology.getPredicate( functor );
         if ( term != null ) {
            return new Belief( term );
         }
      }
      return null;
   }

   public Belief createBelief( String functor, String ontology ) {
      if ( ontologies.contains( ontology ) ) {
         Ontology ont = agent.getOntologyManager().getOntology( ontology );
         FOS term = ont.getPredicate( functor );
         if ( term != null ) {
            return new Belief( term );
         }
      }

      return null;
   }

   public void setActionIdentifier( ActivityIdentifier aid ) {
      this.actionIdentifier = aid;
   }

   public String getArgumentType( int index ) {
      return actionIdentifier.getType( index );
   }

   public Object getArgumentAsObject( FOS action, int index ) {
      Object result = null;
      String type = actionIdentifier.getType( index );
      if ( type.equals( "string" ) ) {
         result = action.argAt( index ).toString();
      }
      else if ( type.equals( "short" ) ) {
         result = Short.parseShort( action.argAt( index ).toString() );
      }
      else if ( type.equals( "int" ) ) {
         result = Integer.parseInt( action.argAt( index ).toString() );
      }
      else if ( type.equals( "long" ) ) {
         result = Long.parseLong( action.argAt( index ).toString() );
      }
      else if ( type.equals( "float" ) ) {
         result = Float.parseFloat( action.argAt( index ).toString() );
      }
      else if ( type.equals( "double" ) ) {
         result = Double.parseDouble( action.argAt( index ).toString() );
      }
      else if ( type.equals( "char" ) ) {
         String s = action.argAt( index ).toString();
         if ( s.length() == 1 ) {
            result = new Character( s.charAt( 0 ) );
         }
         else {
            throw new RuntimeException( "Attempt to convert a string to a char: " + s );
         }
      }
      else {
         Function function = getFunction( type );
         if ( function != null ) {
            result = function.toObject( action.argAt( index ) );
         }
         else {
            // TODO: Think about this - current default action is to return as a
            // string
            result = action.argAt( index ).toString();
         }
      }

      return result;
   }

   private Function getFunction( String type ) {
      for ( String ontologyName : ontologies ) {
         Ontology ontology = agent.getOntologyManager().getOntology( ontologyName );
         for ( Function function : ontology.getFunctions() ) {
            if ( function.getFunctor().equals( type ) ) {
               return function;
            }
         }
      }
      return null;
   }

   /***************************************************************************
    * BELIEF MANIPULATION METHODS
    ***************************************************************************/
   protected void adoptBelief( String bel ) {
      agent.adoptBelief( new FOS( bel ) );
   }

   public void adoptBelief( Belief belief ) {
      agent.adoptBelief( belief.toBelief() );
   }

   protected void retractBelief( String bel ) {
      agent.dropBelief( new FOS( bel ) );
   }

   public String getAgentName() {
      return agent.getName();
   }

   public boolean performAction( FOS action ) {
      //long startTime = System.currentTimeMillis();
      boolean result = act( action );
      //long endTime = System.currentTimeMillis();
      // updatePerformance(endTime-startTime);
      return result;
   }

   public void setConfiguration( Map<String,String> configuration ) {
      this.configuration = configuration;
   }

   public Map<String,String> getConfiguration() {
      return configuration;
   }

   /***************************************************************************
    * ACTION METHOD
    ***************************************************************************/
   public abstract boolean act( FOS action );
}
