package com.agentfactory.logic.reasoner;

import com.agentfactory.logic.lang.FOS;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * 
 * @author Terry Lowen
 * @version July 2002
 */
public class SubstitutionSet implements Iterable<Substitution>, Cloneable {
    private static final Logger logger = Logger.getLogger( SubstitutionSet.class.getName() );

    static {
        logger.setLevel( Level.SEVERE );
    }

    List<Substitution> substitutions;

    public SubstitutionSet() {
        substitutions = new ArrayList<Substitution>();
    }

    /**
     * Return the constant that corresponds to the given variable (if one
     * exists).
     * 
     * @param aVar
     * @return
     */
    public FOS findConstant( FOS aVar ) {
        for ( Substitution subs : substitutions ) {
            if ( subs.getFirst().isSameAs( aVar ) )
                return subs.getSecond();
        }

        return aVar;
    }

    public boolean hasVariable( FOS aVar ) {
        if ( substitutions.isEmpty() ) {
            return false;
        }

        int i = 0;
        boolean found = false;
        while ( !found && ( i < substitutions.size() ) ) {
            found = substitutions.get( i ).getFirst().isSameAs( aVar );
            i++;
        }

        return found;
    }

    @Override
    public boolean equals( Object o ) {
        if ( o instanceof SubstitutionSet ) {
            SubstitutionSet set = (SubstitutionSet) o;
            if ( substitutions.size() != set.substitutions.size() ) {
                return false;
            }

            int i = 0;
            boolean equal = true;
            while ( equal && i < substitutions.size() ) {
                equal = substitutions.get( i ).equals( set.substitutions.get( i ) );
                i++;
            }
            return equal;
        }

        return false;
    }

    /**
     * Generate an Idempotent Binding
     * 
     * @return
     */
    public SubstitutionSet idempotencyBuilder() {
        logger.log( Level.INFO, "Initial Binding: " + this );
        FOS first, second;
        SubstitutionSet subs = new SubstitutionSet();
        for ( Substitution substitution : substitutions ) {
            if ( substitution.getFirst().isVariable() ) {
                first = substitution.getFirst();
            }
            else {
                first = substitution.getFirst().apply( this );
            }

            if ( substitution.getSecond().isVariable() ) {
                second = substitution.getSecond();
            }
            else {
                second = substitution.getSecond().apply( this );
            }
            subs.addBinding( new Substitution( first, second ) );
        }
        logger.log( Level.INFO, "Final Binding: " + subs );
        return subs;
    }

    public SubstitutionSet mostGeneralUnifier() {
        SubstitutionSet subs = new SubstitutionSet();
        for ( Substitution substitution : substitutions ) {
            subs.addBinding( new Substitution( substitution.getFirst(), substitution.getSecond().apply( this ) ) );
        }
        return subs;
    }

    /**
     * Take a substitution set and merge it with the current substitutiion set,
     * resulting in a new substitution set.
     * 
     * @param inSubSet
     * @return a SubstitutionSet
     */
    public SubstitutionSet merge( SubstitutionSet inSubSet ) {
        SubstitutionSet subSet1 = new SubstitutionSet();

        for ( Substitution substitution : substitutions ) {
            subSet1.addBinding( substitution );
        }

        for ( Substitution substitution : inSubSet.substitutions ) {
            subSet1.addBinding( substitution );
        }

        return subSet1.idempotencyBuilder();
    }

    @Override
    public String toString() {
        StringBuffer s = new StringBuffer();
        for ( Substitution substitution : substitutions ) {
            s.append( substitution.toString() + " " );
        }
        return s.toString();
    }

    /* Slightly more convenient to use than |addBinding|. */
    public void bind( FOS v, FOS e ) {
        addBinding( new Substitution( v, e ) );
    }

    public void addBinding( Substitution substitution ) {
        substitutions.add( substitution );
    }

    public Substitution[] bindings() {
        return (Substitution[]) substitutions.toArray();
    }

    public boolean isEmpty() {
        return substitutions.isEmpty();
    }

    public int size() {
        return substitutions.size();
    }

    public Substitution getBinding( int i ) {
        return substitutions.get( i );
    }

    public Iterator<Substitution> iterator() {
        return substitutions.iterator();
    }

    public Substitution removeBinding( int index ) {
        return substitutions.remove( index );
    }

    public Object clone() {
        SubstitutionSet set = new SubstitutionSet();
        for ( Substitution substitution : substitutions ) {
            set.substitutions.add( (Substitution) substitution.clone() );
        }

        return set;
    }
}
