package com.agentfactory.platform.core;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * This class implements the Agent Identifier as is required by the FIPA Standards. Support for
 * converting Strings into AgentID objects is provided through the @link FIPAHelper. 
 * 
 * @author  Rem Collier
 *
 * added to by Donal O'Kane 28/02/05 - for storing a description of the services provided by an agent
 */
public class AgentID implements Serializable,Comparable<AgentID> {
    /**
	 * Required for Serialization
	 */
	private static final long serialVersionUID = -8420532173439214594L;
	
	/**
	 * The unique identifier (name) of the agent
	 */
	private String name;
	
	/**
	 * A list of string URLs representing communication channels through which this agent can be contacted
	 */
    private List<String> addresses;
    
    /**
     * List of string URLs representing resolvers that can be used to retrieve up-to-date addresses for the
     * agent should the list of addresses become invalid
     */
    private List<String> resolvers;
    
    /**
     * list of services provided by an agent - this is not part of the FIPA standards and should really be
     * somewhere else.
     */
    private List<String> servicedescriptions;
    
    /**
     * Creates a new instance of AgentID using only the agent name and the list of communication channel URLs
     * @param name the unique name of the agent
     * @param addresses a list of URLs representing communication channels through which the agent can be contacted  
     */
    public AgentID(String name, List<String> addresses) {
        this(name, addresses, new ArrayList<String>(), new ArrayList<String>());
    }
    
    /**
     * Creates a new instance of AgentID using only the agent name and the list of communication channel URLs
     * @param name the unique name of the agent
     * @param addresses a list of URLs representing communication channels through which the agent can be contacted  
     * @param resolvers list of URLs representing resolvers that can be used to obtain an up-to-date list of addresses 
     */
    public AgentID(String name, List<String> addresses, List<String> resolvers) {
        this(name, addresses, resolvers, new ArrayList<String>());
    }
    
    
    /**
     * Creates a new instance of AgentID using only the agent name and the list of communication channel URLs
     * @param name the unique name of the agent
     * @param addresses a list of URLs representing communication channels through which the agent can be contacted  
     * @param resolvers list of URLs representing resolvers that can be used to obtain an up-to-date list of addresses 
     * @param servicedescriptions list of services offered by the agent???
     */
    public AgentID(String name, List<String> addresses, List<String> resolvers, List<String> servicedescriptions) {
        this.name = name;
        this.addresses = addresses;
        this.resolvers = resolvers;
        this.servicedescriptions = servicedescriptions;
    }    

    public String getName() {
        return name;
    }
    
    public List<String> getAddresses() {
        return addresses;
    }
    
    public List<String> getResolvers() {
        return resolvers;
    }
    
     public List<String> getServiceDescriptionss() {
        return servicedescriptions;
    }
    
    public void setName(String name) {
        this.name = name;
    }
    
    public Object clone() {
        List<String> addr = new LinkedList<String>();
        List<String> res = new LinkedList<String>();
        for (String address:addresses) {
            addr.add(new String(address));
        }
        
        for (String resolver: resolvers) {
            res.add(new String(resolver));
        }
        
        return new AgentID(name, addr, res);
    }
    
    public boolean equals(Object obj) {
    	if (!AgentID.class.isInstance(obj)) return false;
    	
    	AgentID aid = (AgentID) obj;
    	return aid.name.equals(name);
    }

	@Override
	public int compareTo(AgentID aid) {
		return name.compareTo(aid.name);
	}
	
	public int hashCode() {
		return name.hashCode();
	}
	
	/**
	 * Match the agentID to the given agentID. The name, and all addresses / resolvers
	 * in the argument agentID can be regular expressions.
	 * 
	 * The main requirement for success here is that everything in the argument is also
	 * in this agent id. If there is something that is in the argument but not in the
	 * agent id, then the match fails. 
	 * 
	 * @param agentID this is a templated agentID whose content can include regular expressions
	 * @return
	 */
	public boolean matches(AgentID agentID) {
		// First, compare the names. If there is no match, return false
		if (!name.matches(agentID.name)) return false;
			
		// Now compare any provided addresses. If any one of these addresses 
		// does not match, return false
		for (String address : agentID.addresses) {
			boolean match = false;
			for (String addr : addresses) {
				if (addr.matches(address)) {
					match = true;
				}
			}
			if (!match) return false;
		}
		
		// Finally, compare any provided resolvers. Again, if any of these
		// addresses does not match, return false
		for (String address : agentID.resolvers) {
			boolean match = false;
			for (String addr : resolvers) {
				if (addr.matches(address)) {
					match = true;
				}
			}
			if (!match) return false;
		}
		
		return true;
	}
}
