package com.agentfactory.platform.impl;

import com.agentfactory.platform.core.IAgent;
import com.agentfactory.platform.core.IAgentPlatform;
import com.agentfactory.platform.core.IPlatformService;
import com.agentfactory.platform.core.IPlatformServiceManager;
import com.agentfactory.platform.core.NoSuchServiceException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author daithi
 */
public abstract class AbstractPlatformServiceManager extends Observable implements IPlatformServiceManager {
    private static final String LOG_NAME = AbstractPlatformServiceManager.class.getName();
    
    static {
        Logger.getLogger(LOG_NAME).setLevel(Level.WARNING);
    }
    protected List<IPlatformService> services;
    protected Map<String, IPlatformService> map;
    
    public AbstractPlatformServiceManager() {
        this.services = new ArrayList<IPlatformService>();
        this.map = new HashMap<String, IPlatformService>();
    }

    protected List<IPlatformService> getRegisteredServices() {
        return this.services;
    }

    /**
     * Load a static service (i.e. already in the classpath prior to runtime)
     *
     * @param serviceClass Class to instantiate for the service
     * @param serviceName Name of the service
     * @param priority Service Priority
     * @param config The configuration for this service
     */
    public void addService( Class<? extends IPlatformService> serviceClass, String serviceName, int priority, Properties config ) throws NoSuchServiceException {
        try {
            IPlatformService service = (IPlatformService) serviceClass.newInstance();
            service.setPriority( priority );
            service.setName( serviceName );
            service.configure( config, this );
            registerService( service );
            setChanged();
            notifyObservers("registered " + service.getName());
            service.start();
        }
        catch ( InstantiationException ie ) {
            System.err.println( "Failed to Instantiate Service [" + serviceClass.getName() + "]" );
        }
        catch ( IllegalAccessException iae ) {
            System.err.println( "Failed to Instantiate Service [" + serviceClass.getName() + "]" );
        }
    }

    /**
     * Convenience method to add a platform service with default priority and configuration
     * @param serviceClass Class to instantiate for the service
     * @param serviceName The name of the service
     * @throws com.agentfactory.platform.core.NoSuchServiceException if the {@code serviceClass} cannot be instantiated
     */
    public void addService( Class<? extends IPlatformService> serviceClass, String serviceName ) throws NoSuchServiceException {
        this.addService( serviceClass, serviceName, 0, new Properties() );
    }

    public void registerService( IPlatformService service ) {
        if (map.containsKey(service.getName())) {
            Logger.getLogger(LOG_NAME).log(Level.WARNING, "Duplicate Service Registration: " + service.getName());
        } else {
            this.services.add( service );
            this.map.put(service.getName(), service);
        }
    }

    public void unregisterService( IPlatformService service ) {
        this.services.remove( service );
        this.map.remove( service.getName() );
        
        setChanged();
        notifyObservers("unregistered " + service.getName());
    }

    public IPlatformService getServiceByName( String name ) {
        return map.get(name);
    }

    public abstract IAgentPlatform getAgentPlatform();

    public List<IPlatformService> getFilteredServices( Class<? extends IPlatformService> filter ) {
        List<IPlatformService> toReturn = new ArrayList<IPlatformService>();
        
        for ( IPlatformService service : this.getRegisteredServices() ) {
            if ( filter.isInstance( service) ) {
                toReturn.add( service );
            }
        }
        
        return toReturn;
    }
    
    public boolean bindToService(IAgent agent, IPlatformService service) {
        // TODO (Rem Collier): Insert security check here
        service.bind(agent);
        return true;
    }

    public Set<IPlatformService> getServices() {
        return new HashSet<IPlatformService>( this.services );
    }
}
