/**
 * Title:       StringMessage.java
 * Copyright:   Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved.
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */
package com.agentfactory.platform.impl;

import java.util.LinkedList;
import java.util.List;

import com.agentfactory.platform.core.AgentID;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.util.FIPAContent;
import com.agentfactory.platform.util.FIPAHelper;

/**
 * Implementation of FIPA ACL Messaging Representation in String.
 * 
 * @author Rem Collier
 */
public class StringMessage implements IMessage {
    String performative;
    AgentID sender;
    List<AgentID> receivers;
    String content;
    int contentLength;

    String protocol;
    String ontology;
    String language;

    String replyWith;
    String inReplyTo;
    String replyBy;
    AgentID replyTo;
    String conversationId;

    /** Creates a new instance of StringMessage */
    private StringMessage() {
        receivers = new LinkedList<AgentID>();
    }

    public static StringMessage newInstance() {
        return new StringMessage();
    }

    /**
     * Factory method that is responsible for the creation of a StringMessage
     * object based upon the String passed as a parameter.
     * 
     * @param content
     *            the message in a string format
     * @return the message as a StringMessage object
     */
    public static StringMessage newInstance( String content ) {
        StringMessage message = new StringMessage();

        // Convert the string to FIPA content, and then process...
        FIPAContent parts = FIPAContent.newInstance( content );
        message.performative = parts.getCommand().toLowerCase();

        // Temporary variables
        FIPAContent part = null;

        // Now process the remaining tokens
        for ( int i = 0; i < parts.size(); i++ ) {
            part = parts.get( i );
            if ( part.getCommand().equals( ":sender" ) ) {
                message.sender = FIPAHelper.fromFIPASL( parts.get( ++i ) );
            }
            else if ( part.getCommand().equals( ":receiver" ) ) {
                part = parts.get( ++i );
                for ( int j = 0; j < part.size(); j++ ) {
                    message.receivers.add( FIPAHelper.fromFIPASL( part.get( j ) ) );
                }
            }
            else if ( part.getCommand().equals( ":content-length" ) ) {
                message.contentLength = Integer.parseInt( parts.get( ++i ).toString() );
            }
            else if ( part.getCommand().equals( ":reply-with" ) ) {
                message.replyWith = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":reply-to" ) ) {
                message.replyTo = FIPAHelper.fromFIPASL( parts.get( ++i ) );
            }
            else if ( part.getCommand().equals( ":language" ) ) {
                message.language = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":ontology" ) ) {
                message.ontology = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":content" ) ) {
                message.content = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":protocol" ) ) {
                message.protocol = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":conversation-id" ) ) {
                message.conversationId = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":in-reply-to" ) ) {
                message.inReplyTo = parts.get( ++i ).toString();
            }
            else if ( part.getCommand().equals( ":reply-by" ) ) {
                message.replyBy = parts.get( ++i ).toString();
            }
        }

        return message;
    }

    public String toFIPAString() {
        StringBuffer buf = new StringBuffer();
        buf.append( "(" ).append( performative );

        buf.append( " :sender " ).append( FIPAHelper.toFIPAString( sender ) ).append( " :receiver (set" );

        for ( int i = 0; i < receivers.size(); i++ ) {
            buf.append( " " ).append( FIPAHelper.toFIPAString( (AgentID) receivers.get( i ) ) );
        }

        buf.append( ")" );

        if ( contentLength > 0 ) {
            buf.append( " :content-length " ).append( contentLength );
        }

        if ( replyWith != null ) {
            buf.append( " :reply-with " ).append( replyWith );
        }

        if ( inReplyTo != null ) {
            buf.append( " :in-reply-to " ).append( inReplyTo );
        }

        if ( replyTo != null ) {
            buf.append( " :reply-to " ).append( FIPAHelper.toFIPAString( replyTo ) );
        }

        if ( language != null ) {
            buf.append( " :language " ).append( language );
        }

        if ( ontology != null ) {
            buf.append( " :ontology " ).append( ontology );
        }

        if ( content != null ) {
            buf.append( " :content " ).append( content );
        }

        if ( protocol != null ) {
            buf.append( " :protocol " ).append( protocol );
        }

        if ( conversationId != null ) {
            buf.append( " :conversation-id " ).append( conversationId );
        }

        if ( replyBy != null ) {
            buf.append( " :reply-by " ).append( replyBy );
        }

        buf.append( ")" );

        return buf.toString();
    }

    public String toString() {
        return toFIPAString();
    }

    public String getAclRepresentation() {
        return "fipa.acl.rep.string.std";
    }

    public String getPayloadEncoding() {
        return "US-ASCII";
    }

    public void setPerformative( String performative ) {
        this.performative = performative;
    }

    public String getPerformative() {
        return performative;
    }

    public void setSender( AgentID sender ) {
        this.sender = sender;
    }

    public AgentID getSender() {
        return sender;
    }

    public List<AgentID> getReceivers() {
        return receivers;
    }

    public void setContent( String content ) {
        this.content = content;
        contentLength = content.length();
    }

    public String getContent() {
        return content;
    }

    public int getContentLength() {
        return contentLength;
    }

    public void setReplyTo( AgentID replyTo ) {
        this.replyTo = replyTo;
    }

    public AgentID getReplyTo() {
        return replyTo;
    }

    public void setReplyWith( String replyWith ) {
        this.replyWith = replyWith;
    }

    public String getReplyWith() {
        return replyWith;
    }

    public void setReplyBy( String replyBy ) {
        this.replyBy = replyBy;
    }

    public String getReplyBy() {
        return replyBy;
    }

    public void setInReplyTo( String inReplyTo ) {
        this.inReplyTo = inReplyTo;
    }

    public String getInReplyTo() {
        return inReplyTo;
    }

    public void setLanguage( String language ) {
        this.language = language;
    }

    public String getLanguage() {
        return language;
    }

    public void setOntology( String ontology ) {
        this.ontology = ontology;
    }

    public String getOntology() {
        return ontology;
    }

    public void setProtocol( String protocol ) {
        this.protocol = protocol;
    }

    public String getProtocol() {
        return protocol;
    }

    public void setConversationId( String conversationId ) {
        this.conversationId = conversationId;
    }

    public String getConversationId() {
        return conversationId;
    }

    public byte[] toByteArray() {
        return toFIPAString().getBytes();
    }
}
