/**
 * Title:       FIPAContent.java
 * Copyright: 
 * Copyright (c) 1996-2004 The Agent Factory Working Group. All rights reserved.
 * Licence:     This file is free software; you can redistribute it and/or modify
 *              it under the terms of the GNU Lesser General Public License as published by
 *              the Free Software Foundation; either version 2.1, or (at your option)
 *              any later version.
 *
 *              This file is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *              GNU Lesser General Public License for more details.
 *
 *              You should have received a copy of the GNU Lesser General Public License
 *              along with Agent Factory; see the file COPYING.  If not, write to
 *              the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *              Boston, MA 02111-1307, USA.
 */
package com.agentfactory.platform.util;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.agentfactory.platform.core.AgentID;

/**
 * This class contains various methods that support the ALPHAAgent Factort
 * implementation of the FIPA Standards.
 * 
 * @author Rem Collier
 */
public class FIPAHelper {
    /**
     * Creates a new instance of AgentID based upon the FIPA content format. The
     * FIPA content format is: (agent-identifier :name rem :addresses (sequence
     * addr1) :resolvers (sequence addr1))
     * 
     * @param content
     *            the fipa content to be parsed
     */
    public static AgentID fromFIPASL( FIPAContent content ) {
        String name = null;
        List<String> addresses = new LinkedList<String>();
        List<String> resolvers = new LinkedList<String>();
        List<String> servicedescriptions = new LinkedList<String>();
        FIPAContent argument = null;
        if ( content.getCommand().equals( "agent-identifier" ) ) {
            for ( int i = 0; i < content.size(); i++ ) {
                argument = content.get( i );
                if ( argument.getCommand().equals( ":name" ) ) {
                    name = content.get( ++i ).toString();
                }
                else if ( argument.getCommand().equals( ":addresses" ) ) {
                    argument = content.get( ++i );
                    for ( int j = 0; j < argument.size(); j++ ) {
                        addresses.add( argument.get( j ).toString() );
                    }
                }
                else if ( argument.getCommand().equals( ":resolvers" ) ) {
                    argument = content.get( ++i );
                    for ( int j = 0; j < argument.size(); j++ ) {
                        resolvers.add( argument.get( j ).toString() );
                    }
                }
                else if ( argument.getCommand().equals( ":servicedescriptions" ) ) {
                    argument = content.get( ++i );
                    for ( int j = 0; j < argument.size(); j++ ) {
                        servicedescriptions.add( argument.get( j ).toString() );
                    }
                }
            }
        }
        return new AgentID( name, addresses, resolvers, servicedescriptions );
    }

    public static void toXMLString( AgentID agentID, StringBuffer buf, String offset ) {
        buf.append( offset + "<agent-identifier>\n" );
        buf.append( offset + "\t<name>" + agentID.getName() + "</name>\n" );
        buf.append( offset + "\t<addresses>\n" );
        for ( String address : agentID.getAddresses() ) {
            buf.append( offset + "\t\t<url>" + address + "</url>\n" );
        }
        buf.append( offset + "\t</addresses>\n" );
        buf.append( offset + "</agent-identifier>\n" );
    }

    public static String toFIPAString( AgentID agentID ) {
        StringBuffer buf = new StringBuffer();
        buf.append( "(agent-identifier " ).append( ":name " ).append( agentID.getName() ).append( " :addresses " ).append( "(sequence" );

        for ( String address : agentID.getAddresses() ) {
            buf.append( " " ).append( address );
        }

        buf.append( "))" );

        return buf.toString();
    }

    public static String toFOSString( AgentID agentID ) {
        StringBuffer buf = new StringBuffer();
        buf.append( "agentID(" ).append( agentID.getName() ).append( ",addresses(" );

        boolean first = true;
        for ( String address : agentID.getAddresses() ) {
            if ( first ) {
                first = false;
            }
            else {
                buf.append( "," );
            }

            buf.append( "\"" + address + "\"" );
        }

        buf.append( "))" );

        return buf.toString();
    }

    /**
     * This method tokenizes a FIPA Message String delimited by spaces that are
     * not wrapped in brackets.
     * 
     * E.g. (example token (another token)) ^ ^ ^ Delimiters Not a Delimiter
     * 
     * @param message
     *            the FIPA message to be tokenized
     * @return an arraylist of tokens
     */
    public static List<String> getMessageTokens( String message ) {
        List<String> tokens = new ArrayList<String>();

        int length = message.length();
        // if (length == 0) return tokens;

        // Start by finding the opening bracket
        int index = 0;
        while ( ( message.charAt( index ) != '(' ) && ( index < length - 1 ) ) {
            index++;
        }
        int lastIndex = ++index;

        // now extract the "command" part (<commmand> [<type> <param> ]*)
        // this can be detected by either a space, colon, or open bracket
        boolean found = false;
        String tok = null;
        while ( !found && ( index < length ) ) {
            found = ( message.charAt( index ) == '(' ) || ( message.charAt( index ) == ':' ) || ( message.charAt( index ) == ' ' );
            index++;
        }
        if ( found ) {
            tok = message.substring( lastIndex, index - 1 ).trim();
            if ( tok.length() > 0 ) {
                tokens.add( tok );
            }
            lastIndex = index - 1;
            index = index - 1;
            tok = null;
        }

        // now start work on the parameters
        int count = 0;
        boolean quotes = false;
        while ( ( count > -1 ) && ( index < length ) ) {
            if ( message.charAt( index ) == '\"' ) {
                if ( quotes ) {
                    quotes = false;
                }
                else {
                    quotes = true;
                }
            }
            else if ( message.charAt( index ) == '(' ) {
                count++;
            }
            else if ( message.charAt( index ) == ')' ) {
                if ( ( count == 1 ) && !quotes ) {
                    tok = message.substring( lastIndex, index + 1 ).trim();
                    lastIndex = index + 1;
                    count = 0;
                }
                else {
                    count--;
                }
            }
            else if ( ( message.charAt( index ) == ' ' ) && ( count == 0 ) && !quotes ) {
                tok = message.substring( lastIndex, index ).trim();
                lastIndex = index + 1;
            }

            if ( tok != null ) {
                if ( tok.length() > 0 ) {
                    tok = strip( tok, "\"" );
                    tok = strip( tok, "\\" );
                    tokens.add( tok );
                    tok = null;
                }
            }
            index++;
        }

        // Add the remaining token
        if ( ( lastIndex > -1 ) && ( lastIndex < length ) ) {
            tok = message.substring( lastIndex, index - 1 ).trim();
            if ( tok.length() > 0 )
                tokens.add( tok );
        }

        // What about if no tokens?
        if ( tokens.size() == 0 ) {
            tokens.add( message );
        }

        return tokens;
    }

    private static String strip( String text, String ch ) {
        String temp = text;

        int index = temp.indexOf( ch );
        while ( index >= 0 ) {
            temp = temp.substring( 0, index ) + temp.substring( index + 1 );
            index = temp.indexOf( ch );
        }

        return temp;
    }
}
