package com.agentfactory.rma;
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
import com.agentfactory.platform.core.AgentID;
import com.agentfactory.platform.core.IMessage;
import com.agentfactory.platform.core.IPlatformService;
import com.agentfactory.platform.impl.StringMessage;

/**
 * This is an abstract base class the implements the core functionality
 * required by the message handling architecture that is used as part of
 * the ReactingMessageAgent.
 * 
 * Actual message handlers must implement two methods:
 * 
 * <ul>
 * <li> The filter(...) method is used to check whether or not the message
 * handler should be used to attempt to handle the current message.
 * <li> The act(...) method is used to implement the handling of the message.
 * </ul>
 * 
 * Additionally, the send(...) and getService(...) methods are provided as helper
 * methods to simplify the implements of the handlers.
 * 
 * @author remcollier
 */
public abstract class MessageHandler {
    /**
     * The agent that the handler is used by
     */
    protected ReactiveMessageAgent agent;
    
    /**
     * The performative of the type of message that the handler can handle.
     */
    private String performative;
    
    /**
     * Checks whether or not the message can be handled by the handler.
     * 
     * @param message the message to be handled
     * @return true if the message can be handled by the handler, false otherwise.
     */
    public abstract boolean filter(IMessage message);
    
    /**
     * The actual implementation of the message handler.
     * 
     * @param message the message to be handled.
     * @return true if the message was handled, false otherwise.
     */
    public abstract boolean act(IMessage message);
    
    /**
     * Constructor that is used to set up initial fields in the handler.
     * 
     * @param agent the agent that the handler is associated with
     * @param performative the performative for the hander
     */
    protected MessageHandler(ReactiveMessageAgent agent, String performative) {
        this.agent = agent;
        this.performative = performative;
    }
    
    /**
     * Return the performative for the handler
     * 
     * @return a String that represents the performative
     */
    public String getPerformative() {
        return performative;
    }
        
    /**
     * Helper method that supports the sending of messages by a handler
     * 
     * @param message the message to be sent
     * @return true if the message was sent succesfully, false otherwise.
     */
    public boolean send(IMessage message) {
        return agent.send(message);
    }
    
    /**
     * Helper method that returns the platform service for a given id
     * 
     * @param id the id of the platform service
     * @return a reference to the platform service or null
     */
    public IPlatformService getService(String id) {
        return agent.getPlatformService(id);
    }
    
    public IMessage createAFAPLMessage(String performative, AgentID sender, IMessage message, String content) {
        StringMessage response = StringMessage.newInstance();
        response.setLanguage("AFAPL");
        response.setSender(sender);
        response.getReceivers().add(message.getSender());
        response.setContent(content);
        response.setPerformative(performative);
        response.setConversationId(message.getConversationId());
        return response;
    }
}
