package com.agentfactory.visualiser;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.List;
import java.util.Observable;
import java.util.Observer;
import java.util.Set;

import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.WindowConstants;

import com.agentfactory.platform.core.AgentID;
import com.agentfactory.visualiser.impl.MessageStore;


/**
 * Draws a protocol-diagram like graph based on agent interactions (messages).
 * 
 * @author Zhenhua Yang
 */
public class ProtocolFrame extends JInternalFrame implements Observer {
	
	private static final int INITIAL_POSITION_X = 20;
	private static final int INITIAL_POSITION_Y = 20;
	
	private static final int AGENT_BOX_HEIGHT = 20;
	
	private static final int AGENT_BOX_DISTANCE = 40;
	private static final int TEXT_MARGIN = 4;
	private static final int VERTICAL_LINE_HEIGHT = 30;
	private static final int VERTICAL_BAR_WIDTH = 4;
	private static final int VERTICAL_BAR_HEIGHT = 40;
	private static final int VERTICAL_MESSAGE_MARGIN = 4;
	private static final int TAIL_VERTICAL_LINE_HEIGHT = 50;
	
	/*
	 * The mapping between a Agent and its X position on the graph 
	 */
	private HashMap<AgentID, Integer> agentBoxWidth;
	private Set<AgentID> allAgentIDs;
	private List<InteractionRecord> allInteractionRecords;
	
	private Set<AgentID> displayAgentIDs;
	private List<InteractionRecord> displayInteractionRecords;
	
	private ProtocolPanel diagramPanel;
	private MessageStore messageStore;
	private FilterPanel filterPanel;
	private boolean first = true;
	
	public ProtocolFrame(MessageStore messageStore) {
		this.messageStore = messageStore;
		init();
	}

	private void init() {
		agentBoxWidth = new HashMap<AgentID, Integer>();
		allAgentIDs = messageStore.getAllAIDs();
		allInteractionRecords = messageStore.getInteractionRecords();
		
		displayAgentIDs = allAgentIDs;
		displayInteractionRecords = allInteractionRecords;
		
		filterPanel = new FilterPanel(allAgentIDs);
		filterPanel.setEventHandler(new MyButtonClickProcessor());
		
		buildUI();
	}

	private void buildUI() {
		getContentPane().removeAll();
        setTitle("AF Protocol Viewer");
        this.setResizable(true);
        this.setClosable(true);
        this.setMaximizable(true);

        setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        getContentPane().setLayout(new BorderLayout());

        diagramPanel = new ProtocolPanel();
        
        JScrollPane scroll = new JScrollPane(diagramPanel);
        scroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
        scroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        
        JSplitPane pane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        pane.setRightComponent(scroll);
        
        filterPanel.setPreferredSize(new Dimension(200, 600));
        pane.setLeftComponent(filterPanel);
        pane.setDividerLocation(0.9);
        getContentPane().add(pane, BorderLayout.CENTER);
        validate();
		if (first) {
	        setSize(800, 600);
	        first = false;
		}
        setVisible(true);
	}
	
	private void applyFilter(Set<AgentID> filterAIDSet) {
		displayAgentIDs = filterAIDSet;
		displayInteractionRecords = new Filter(filterAIDSet).filterInteractionRecords(allInteractionRecords);
	}
	
	/**
     * handles button click events for all buttons
     */
    class MyButtonClickProcessor implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            try {
                if (e.getActionCommand().equalsIgnoreCase("Update Filter")) {
                	applyFilter(filterPanel.getSelectedAIDs());
                	buildUI();
                } 
            } catch (Exception ex) {
                ex.printStackTrace();
            }
        }
    }
	
	/**
	 * The actual canvas for drawing the graph
	 */
	class ProtocolPanel extends JPanel {
		
		int characterWidth;
		
		private int getStringWidth(String str) {
			return characterWidth * str.length();
		}
		
		public void paintComponent(Graphics g) {
			super.paintComponent(g);
			
			if (displayAgentIDs == null || displayAgentIDs.size()<2) return;
			
			diagramPanel.setBackground(Color.WHITE);
			
			characterWidth = MiscUtils.getAverageCharacterWidthOnGraph(g);
			
			int xx = drawAgents(g, displayAgentIDs);
			
			int yy = drawInteractions(g, displayInteractionRecords);
			
			setPreferredSize(new Dimension(xx + 20, yy));
		}

		private int drawAgents(Graphics g, Set<AgentID> agents) {
			int x = INITIAL_POSITION_X;
			int y = INITIAL_POSITION_Y;
			for (AgentID aid : agents) {
				int boxWidth = drawAgentBox(x, y, g, aid);
				x += (boxWidth + AGENT_BOX_DISTANCE);
			}
			return x;
		}
		
		private int drawAgentBox(int x, int y, Graphics g, AgentID aid) {
			int boxWidth = getStringWidth(aid.getName()) + TEXT_MARGIN * 2;
			agentBoxWidth.put(aid, new Integer(boxWidth));
			
			g.drawRect(x, y, boxWidth, AGENT_BOX_HEIGHT);
			g.drawString(aid.getName(), x + TEXT_MARGIN, y + AGENT_BOX_HEIGHT - TEXT_MARGIN);
			
			return boxWidth;
		}
		
		// draw interaction messages and lines
		private int drawInteractions(Graphics g, List<InteractionRecord> msgs) {
			int y = INITIAL_POSITION_Y + AGENT_BOX_HEIGHT;
			
			for (InteractionRecord record : msgs) {
				int x = getStartPosX(record.getSenderSimpleAID());
				y = drawInteraction(x, y, g, record);
			}
			
			drawTailVerticalLines(y, g);
			return y + TAIL_VERTICAL_LINE_HEIGHT;
		}

		private int drawInteraction(int x, int y, Graphics g, InteractionRecord record) {
			int receiverX = getStartPosX(record.getReceiverSimpleAIDs().get(0));
			int y2 = drawVerticalLines(x, receiverX, y, g);
			int y3 = drawVerticalBars(x, receiverX, y2, g);
			drawMessage(x, receiverX, y2, g, record);
			
			// draw vertical lines for agents not involved in the interaction
			for (AgentID aid : displayAgentIDs) {
				if (aid.equals(record.getSenderSimpleAID()) || record.getReceiverSimpleAIDs().contains(aid)) {
					continue;
				} else {
					int otherAgentX = getStartPosX(aid);
					drawVerticalLine(otherAgentX, y, VERTICAL_LINE_HEIGHT + VERTICAL_BAR_HEIGHT, g);
				}
			}
			return y3;
		}

		private void drawMessage(int senderX, int receiverX, int y, Graphics g, InteractionRecord record) {
			if (record.getStringMessage().getPerformative().equalsIgnoreCase("request")) {
				g.setColor(Color.RED);
			} else {
				g.setColor(Color.BLUE);
			}
			// draw horizontal line
			int lineStartX = receiverX > senderX ? senderX + VERTICAL_BAR_WIDTH / 2 : senderX - VERTICAL_BAR_WIDTH / 2;
			int lineEndX = receiverX > senderX ? receiverX - VERTICAL_BAR_WIDTH / 2 : receiverX + VERTICAL_BAR_WIDTH / 2;
			int lineStartY = y + VERTICAL_MESSAGE_MARGIN;
			int lineEndY = lineStartY;
			g.drawLine(lineStartX, lineStartY, lineEndX, lineEndY);
			
			// draw arrow
			int arrowX = receiverX > senderX ? lineEndX - 6 : lineEndX + 6;
			g.drawLine(arrowX, lineEndY - 4, lineEndX, lineEndY);
			g.drawLine(arrowX, lineEndY + 4, lineEndX, lineEndY);
			
			//draw message
			int msgStartX = receiverX > senderX ? lineStartX + 15 : lineEndX + 15;
			int msgStartY = lineStartY - 8;
			String msgText = getMessageTextForDiagram(record);
			g.drawString(msgText, msgStartX, msgStartY);
			
			g.setColor(Color.BLACK);
		}
		
		private String getMessageTextForDiagram(InteractionRecord record) {
			String performative = record.getStringMessage().getPerformative().toLowerCase();
			String content = record.getStringMessage().getContent();
			if (content.length() > 20) {
				content = content.substring(0, 20) + "...";
			}
			return  performative + "(" + content + ")";
		}

		private int drawVerticalLines(int senderX, int receiverX, int y, Graphics g) {
			drawVerticalLine(senderX, y, VERTICAL_LINE_HEIGHT, g);
			drawVerticalLine(receiverX, y, VERTICAL_LINE_HEIGHT, g);
			
			return y + VERTICAL_LINE_HEIGHT;
		}

		private int drawVerticalBars(int senderX, int receiverX, int y, Graphics g) {
			drawVerticalBox(senderX, y, VERTICAL_BAR_HEIGHT, g);
			drawVerticalBox(receiverX, y, VERTICAL_BAR_HEIGHT, g);

			return y + VERTICAL_BAR_HEIGHT;
		}

		private void drawVerticalBox(int x, int y, int height, Graphics g) {
			g.drawRect(x - VERTICAL_BAR_WIDTH / 2, y, VERTICAL_BAR_WIDTH, height);
		}

		private void drawVerticalLine(int x, int y, int height, Graphics g) {
			g.drawLine(x, y, x, y + height);
		}
		
		private void drawTailVerticalLines(int y, Graphics g) {
			for (AgentID aid : displayAgentIDs) {
				int x = getStartPosX(aid);
				g.drawLine(x, y, x, y + TAIL_VERTICAL_LINE_HEIGHT);
			}
		}

		// start position x for vertical line
		private int getStartPosX(AgentID senderSimpleAID) {
			
			int x = INITIAL_POSITION_X;
			
			Object[] objs = displayAgentIDs.toArray();
			
			for (int i = 0; i< objs.length; i++) {
				AgentID aid = (AgentID) objs[i];
				if (aid.equals(senderSimpleAID)) {
					x += (agentBoxWidth.get(aid).intValue() / 2);
					break;
				} else {
					int boxWidth = agentBoxWidth.get(aid).intValue();
					x += boxWidth + AGENT_BOX_DISTANCE;
				}
			}
			return x;
		}
	}

	@Override
	public void update(Observable o, Object arg) {
		init();
	}
}
